#!/usr/bin/env python


"""
    ### Whois: Utilities

    Here is the list of general functions used for whois queries
    for IP addresses and domains
"""

import ipaddress


# validate the Ipv4 or IPv6
# if the IP address is valid, returns IP version (4 or 6), otherwise False
def validate_ip(ip):
    try:
        ip = ipaddress.ip_address(ip)
        return str(ip.version)
    except ValueError:
        return False



#!/usr/bin/env python

from bs4 import BeautifulSoup
from urllib.parse import urlparse
from typing import Dict, Any

from config import config
from modules.utilities.url_sanitizer import url_sanitizer


def link_finder(soup: BeautifulSoup, domain: str) -> Dict[str, Any]:
    """
    This function looks for any links within the HTML page. It would checks if
    the link is internal or external and write them in the specified dictionary.

    Args:
        soup (BeautifulSoup): the BeautifulSoup format of the HTML page
        domain (str): the domain name

    Returns:
        links (dict): a dictionary contains all (external and internal) links
    """
    # a temporary variable
    temp = []
    # a variable to store results in
    links = {
        'internal': [],
        'external': [],
        'count': {
            'internal': 0,
            'external': 0,
        }
    }

    # Initialize element counts to 0
    for element_link in config['html']['elements_links']:
        links['count'][element_link[0]] = 0
        links['count'][f'int_{element_link[0]}'] = 0
        links['count'][f'ext_{element_link[0]}'] = 0

    # Find 'link' elements within the HTML content
    for element_link in config['html']['elements_links']:
        elements = soup.findAll(element_link[0])

        for element in elements:
            # Ignore the situation that the HTML element doesn't contain any link
            # e.g., <script>console.log('Hi!')</script>
            if not element.has_attr(element_link[1]):
                continue

            # Define link
            link = element[element_link[1]]

            # Ignore empty links or non-links like anchor, 'mailto', etc.
            if (
                link in temp
                or not link
                or link.startswith(tuple(config['html']['elements_links_ignore']))
            ):
                continue

            # Add the link to the temp in order to prevent duplication
            temp.append(link)

            # Check if the link is external
            # It looks for any scheme like http(s), (s)ftp, sip, etc.
            if (
                urlparse(link).hostname
                and urlparse(link).hostname.replace('www.', '') != url_sanitizer(domain)[1]
            ):
                links['external'].append({
                    'link': link,
                    'html_element': element_link[0]
                })
                # Count the total number of external links and per element
                links['count']['external'] += 1
                links['count'][element_link[0]] += 1
                links['count'][f'ext_{element_link[0]}'] += 1

            # If the link is internal
            else:
                links['internal'].append({
                    'link': link,
                    'html_element': element_link[0]
                })
                # Count the total number of internal links and per element
                links['count']['internal'] += 1
                links['count'][element_link[0]] += 1
                links['count'][f'int_{element_link[0]}'] += 1

    return links



#!/usr/bin/env python

from typing import List, Dict
from bs4 import BeautifulSoup


def form_finder(soup: BeautifulSoup) -> List[Dict[str, any]]:
    """
    This function looks for any forms within the HTML page including all fields.

    Args:
        soup (BeautifulSoup): The BeautifulSoup format of the HTML page

    Returns:
        List[Dict[str, any]]: A list containing all forms with their fields
    """
    forms = []

    # find 'form' elements within the HTML content
    all_forms = soup.find_all('form')

    # iterate over forms in the HTML page
    for form in all_forms:
        # a variable to store inputs of the form
        fields = []

        # get the inputs
        inputs = form.find_all('input')

        # iterate over fields ('input' elements) in the form
        if inputs:
            for input in inputs:
                fields.append({
                    'type': input.get('type', ''),
                    'name': input.get('name', ''),
                    'id': input.get('id', ''),
                    'value': input.get('value', ''),
                    'placeholder': input.get('placeholder', '')
                })

        # write findings in the list
        forms.append({
            'method': form.get('method', ''),
            'action': form.get('action', ''),
            'name': form.get('name', ''),
            'id': form.get('id', ''),
            'fields': fields
        })

    # return results
    return forms



#!/usr/bin/env python

from colorama import Fore, Back, Style
from time import sleep

from config import config
from modules.dns.dns_dig import dns_dig
from modules.dns.google_dns import google_dns
from modules.dns.dns_spy import dns_spy
from modules.whois.ip_api import validate_ip
from modules.whois.ip_api import ip_api
from modules.utilities.printer import printer


# resolve the DNS and get all the defined records
def dns_checker(domain):
    """
        ### DNS Checker

        This function is responsible for getting SSL certificate details through
        different methods. Therefore, it will print the summary results on STDOUT.

        # Input:  - a single domain name
        # Output: - a dictionary contains SSL certificate details
    """
    # a dictionary to store data
    dns_records = dict()
    # get the list of desired DNS records; remove duplicates and sort it
    dns_records_list = sorted(set(config['dns']['dns_records']))

    # print the title of the section
    printer('      │\n      ├───' + Fore.BLACK + Back.WHITE + ' DNS Records ' + Style.RESET_ALL)

    # get data from DNS Spy or return an empty dict
    printer('      │        └□ ' + Fore.GREEN + 'DNS Spy API is calling' + Style.RESET_ALL)
    records, subdomains = dns_spy(domain)

    # iterate over DNS records and resolve them
    for record in dns_records_list:
        # a list to store data
        dns_records[record] = list()

        # load data from built-in python module
        answers = dns_dig(domain, '', record)
        # update it by adding data from Google DNS API
        answers.update(google_dns(domain, '', record))

        # if it is a TXT record
        if record == 'TXT':
            # iterate over the list of subdomains given in the config.py file
            # it can be a brute-force to find more
            for subdomain in sorted(set(config['dns']['txt_records_helper'])):
                # update data from built-in python module for subdomains
                answers.update(dns_dig(subdomain + domain, subdomain, record))
                # update it by adding data from Google DNS API for subdomains
                answers.update(google_dns(subdomain + domain, subdomain, record))

        # print the title of the subsection (DNS Record Name)
        printer('      │      ' + Fore.CYAN + f'■ {record} Records       ' + Style.RESET_ALL)

        # iterate over the DNS records and print the result on STDOUT
        for answer in answers:
            i += 1

            # if it is a TXT record but not listed in the config file
            text_records = config['dns']['include_txt_records']
            if record == 'TXT' and not any(txt.lower() in answer.lower() for txt in text_records):
                color = Fore.MAGENTA
            else:
                color = Fore.YELLOW

            # print the result on STDOUT
            l_start = '       │'  # for printing IP lookups
            if i == len(answers):
                printer('      │      └◌ ' + color + answer['value'] + Style.RESET_ALL)
            else:
                printer('      │      ├◌ ' + color + answer + Style.RESET_ALL)

            # if the result is an IP, print the IP lookup results
            if validate_ip(answer):
                lookup = ip_api(answer)
                if config['verbosity'] >= 3:
                    printer(l_start + ' ' * 14 + '├□ Location:    ' + Fore.CYAN +
                            lookup['city'] + ', ' + lookup['country'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '├□ ISP Name:    ' + Fore.CYAN +
                            lookup['isp'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '├□ Org. Name:   ' + Fore.CYAN +
                            lookup['organization'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '├□ Reverse DNS: ' + Fore.CYAN +
                            lookup['reverse_dns'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '├□ Is Mobile?   ' + Fore.CYAN +
                            lookup['mobile'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '├□ Is Proxy?    ' + Fore.CYAN +
                            lookup['proxy'] + Style.RESET_ALL)
                    printer(l_start + ' ' * 14 + '└□ Is Hosting?  ' + Fore.CYAN +
                            lookup['hosting'] + Style.RESET_ALL)

        # add a delay between DNS queries
        sleep(config['delay']['dns'])

    # return the list of the DNS records
    return dns_records



#!/usr/bin/env python


"""
    ### Output: Utilities

    Here is the list of general functions used for generating output files
"""

from typing import Dict

from config import config


def flatten_json(y):
    out = {}

    def flatten(x, name=''):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + '_')
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + '_')
                i += 1
        else:
            out[name[:-1]] = x

    flatten(y)
    return out


def flatten_json(init_key: str, input: Dict, output: Dict) -> Dict:
    """
    Flatten a nested JSON object into a flat dictionary.
    i.e. {'a': {'b': 1, 'c': {'d': 2}}, 'e': 3} ===> {'a_b': 1, 'a_c_d': 2, 'e': 3}

    Args:
        init_key (str): The initial key to use when flattening the JSON object.
        input (Dict): The JSON object to flatten.
        output (Dict): The dictionary to store the flattened JSON.

    Returns:
        A dictionary containing the flattened JSON object.
    """
    joiner = config['output']['csv']['header_joiner']

    # iterate over keys & values recursively
    for key, value in input.items():
        # check if the json contains another jason (is nested)
        if isinstance(value, dict):
            # run the function recursively
            flatten_json(init_key + key + joiner, value, output)
        else:
            output[init_key + key] = value
    return output

# make the json flat
# i.e. {'a': {'b': 1, 'c': {'d': 2}}, 'e': 3}
# ===> {'a_b': 1, 'a_c_d': 2, 'e': 3}
def flatten_json_old(init_key, json, results):
    joiner = config['output']['csv']['header_joiner']
    # iterate over keys & values recursively
    for key, value in json.items():
        # check if the json contains another jason (is nested)
        if isinstance(value, dict):
            init_key += key + joiner
            # run the function recursively
            flatten_json_old(init_key, value, results)
            # sanitize the aggregated key
            init_key = init_key[:-len(key + joiner)]
        else:
            init_key += key
            results[init_key] = value
            # sanitize the aggregated key
            init_key = init_key[:-len(key)]
    # return results
    return results


# it will get the json, make it flat, and return keys and values in separate lists
# it also checks if it is asked (in the config file) to be in the CSV file or not
# if not, it will be ignored from the output lists
def csv_maker(json, config_json):
    keys = set()
    values = list()
    both = dict()

    #
    delimiter = config['output']['csv']['delimiter']['other']

    # make the json flat
    json = flat_json('', json, {})
    if config_json:
        config_json = flat_json('', config_json, {})

    # print(json.keys(), '\n\n')
    # print(config_json.keys(), '\n\n')
    # iterate over keys & values
    if isinstance(json, dict):
        # iterate over the flatted JSON
        for key, value in json.items():
            # check if it is True in config file to be written in CSV or not
            # additionally ignore the ones that are not in the config
            if key in config_json and config_json[key]:
                print('---------6---------', key)

            if key not in config_json:
                print('---------1---------', key)

            if isinstance(value, list):
                if isinstance(value[0], dict):
                    continue
                value = delimiter.join(value)

            keys.add(key)
            values.append(str(value))
            both[key] = str(value)
    else:
        values = json

    print('+++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    # return results
    return [
        keys,
        values,
        both
    ]



#!/usr/bin/env python

import re

import global_variable


def printer(term: str) -> None:
    """
    This function writes the input term on STDOUT, then removes any ANSI escape
    sequences (sanitization) and keeps the result to be written in a text file.

    Args:
        term: The text that should be written on STDOUT and in the text file
    """
    global ALL_PRINTS

    # Print the term on the terminal (STDOUT)
    print(term)

    # Remove ANSI escape sequences from the term to be able to save it into a file
    term = re.sub(r'\x1B\[\d+m', '', term)

    # Save the term in a global variable to save in a file later
    global_variable.ALL_PRINTS += term + '\n'



#!/usr/bin/env python


"""
    ### Output: CSV

    This function generates a CSV file based on the data

    # Input:  - a filename
              - a dictionary contains all the data
              - a list of tags define what data should be written into the CSV file
                [ssl, whois, http, dns, related_sub_domains]
    # Output: - a CSV file
"""

import csv

def csv_writer(json_data):
    # create a CSV writer object
    with open('1.csv', mode='w', encoding='UTF8', newline='') as output_file:
        writer = csv.writer(output_file, quoting=csv.QUOTE_MINIMAL)



        # start the recursive flattening process
        flatten_json(json_data)



# import csv
# from colorama import Back, Fore, Style

# from config import config
# from modules.output.utils import csv_maker, flat_json
# from modules.utilities.printer import printer
# from modules.utilities.exception_handling import exception_details


# def csv_writer(file_name, json_data, tags):
#     # some variables
#     ssl_header = list()
#     whois_header = list()
#     http_header = set()
#     dns_records = list()
#     related_sub_domains = list()

#     # write the result in the output file
#     try:
#         # print the name of the domain
#         printer(f'\n [{Fore.GREEN}+{Fore.WHITE}]──┬──{Fore.RED}{Back.WHITE} Write into '
#                 f'{file_name}{Style.RESET_ALL}{Fore.GREEN}{Style.RESET_ALL}\n      │')

#         # open the file and start writing in it
#         with open(file_name, mode='w', encoding='UTF8', newline='') as output_file:
#             writer = csv.writer(
#                 output_file, delimiter=config['output']['csv']['delimiter']['column'],
#                 quoting=csv.QUOTE_MINIMAL
#             )

#             # iterate over domains
#             # for domain, results in json_data.items():
#             #     update all HTTP headers
#             #     if tags[2]:
#             #         http_header.update(
#             #             csv_maker(config['output']['include']['http'],
#             #                       config['output']['include']['http'])[0]
#             #         )
#             #     # # formulate the header and write it into the CSV
#             #     if tags[0]:
#             #         ssl_header = csv_maker(
#             #             config['output']['include']['ssl'],
#             #             config['output']['include']['ssl']
#             #         )[0]
#             #     if tags[1]:
#             #         whois_header = csv_maker(
#             #             config['output']['include']['whois'],
#             #             config['output']['include']['whois']
#             #         )[0]
#             #     if tags[2]:
#             #         http_header = csv_maker(
#             #             config['output']['include']['http'],
#             #             config['output']['include']['http']
#             #         )[0]
#             #     if tags[3]:
#             #         dns_records = config['dns']['dns_records'].copy()
#             #     if tags[4]:
#             #         related_sub_domains = ['related_domains', 'subdomains']

#             # form the header
#             header = ['domain'] + ssl_header + dns_records + whois_header + list(http_header) + related_sub_domains

#             # write the header
#             writer.writerow(header)

#             # iterate over domains
#             for domain, results in json_data.items():
#                 # compose the line
#                 line = []
#                 f = ''

#                 # append the domain name to the line
#                 line.append(domain)

#                 # # append the SSL data
#                 # if tags[0]:
#                 #     line += csv_maker(
#                 #         json_data[domain]['ssl'],
#                 #         config['output']['include']['ssl']
#                 #     )[1]

#                 # # append the DNS records data
#                 # if tags[3]:
#                 #     # append DNS records to the line
#                 #     for r in config['dns']['dns_records']:
#                 #         # ignore the TXT details that are defined in the config file
#                 #         # as all will be stored in the key 'txt'
#                 #         line.append(config['output']['csv']['delimiter']['dns_records'].join(results['dns_records'][r]))

#                 # # append the whois data
#                 # if tags[1]:
#                 #     line += csv_maker(
#                 #         json_data[domain]['whois'],
#                 #         config['output']['include']['whois']
#                 #     )[1]

#                 # append the HTTP data
#                 if tags[2]:
#                     # print(results['http'])
#                     f1 = csv_maker(results['http'],
#                                    config['output']['include']['http'])[2]

#                     i = 0
#                     for key in http_header:
#                         if 'redirects' in key:
#                             print('=========', key, f1)
#                         i += 1
#                         if key in f1:
#                             print(f'--{i}-', key,'----    ', f1[key])
#                             line.append(f1[key])
#                         else:
#                             print(f'+-{i}-', key,'----    ')
#                             line.append('')
#                     # line += f
#                     # print(line)

#                 # # append the related domains and subdomains
#                 # if tags[4]:
#                 #     line += csv_maker(
#                 #         json_data[domain]['related_domain'],
#                 #         's'
#                 #     )[1]
#                 #     line += csv_maker(
#                 #         json_data[domain]['subdomain'],
#                 #         ''
#                 #     )[1]

#                 # write the line into the CSV file
#                 writer.writerow(line)

#     except IOError:
#         ex = exception_details()
#         printer(f'      │ ■ {Fore.RED}Error in in I/O (input/output).{Style.RESET_ALL}')
#         if config['verbosity'] >= 4:
#             printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
#                     f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
#     except csv.Error:
#         ex = exception_details()
#         printer(f'      │ ■ {Fore.RED}Error in the CSV module.{Style.RESET_ALL}')
#         if config['verbosity'] >= 4:
#             printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
#     except ValueError:
#         ex = exception_details()
#         printer(f'      │ ■ {Fore.RED}Error in values are going to be written in the CSV file.{Style.RESET_ALL}')
#         if config['verbosity'] >= 4:
#             printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
#     except Exception:
#         ex = exception_details()
#         printer(f'      │ ■ {Fore.RED}Error in writing into the CSV file.{Style.RESET_ALL}')
#         if config['verbosity'] >= 4:
#             printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
#     finally:
#         # print a message shows the scan for the domain is finished
#         printer(f'      │\n      └──{Fore.RED}{Back.WHITE} Finish Writing in "{file_name}" {Style.RESET_ALL}\n')



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Maltiverse API

    This function gets the list of subdomains and corresponding IPs by calling
    the Maltiverse API.

    Read more: https://app.swaggerhub.com/apis-docs/maltiverse/api/1.1

    # Input:  - a single domain name or IP address
              - the type of the asset that needs to be assessed
    # Output: - a set contains vulnerability assessment of the given domain
              - a set of dictionaries contains blocked lists
"""


from config import config
from modules.utils import url_opener, error_printer
from modules.whois.utils import validate_ip


def maltiverse(bacon, type):
    # variables to store subdomains and related domains
    blocked_list = set()
    vulnerabilities = set()

    # the HTTP headers to be sent to Maltiverse
    headers = {'Accept': 'application/json'}

    # form the URL based on the type of the bacon
    if type.lower() == 'ip':
        url = config['api']['maltiverse']['url_ip'].format(bacon)
    if type.lower() == 'domain':
        url = config['api']['maltiverse']['url_domain'].format(bacon)

    # get the results in JSON from Maltiverse
    results = url_opener('GET', url, '', '', headers, 'json', 'Maltiverse API')[0]

    # if the API call returns data
    if not results['status']:
        # iterate over the
        for listed in results['blacklist']:
            blocked_list.add({
                'by': listed['CIArmy'],
                'reason': listed['description'],
                'type': validate_ip(bacon) if validate_ip(bacon) else 'Domain',
                'name': bacon
            })

        # get the vulnerability data based on the type
        if type.lower() == 'domain':
            vulnerabilities.add(
                {
                    'name': 'is_iot_threat',
                    'risk': None if results['is_iot_threat'] else 'high'
                },
                {
                    'name': 'is_cnc',
                    'risk': None if results['is_cnc'] else 'critical'
                },
                {
                    'name': 'is_distributing_malware',
                    'risk': None if results['is_distributing_malware'] else 'critical'
                },
                {
                    'name': 'is_mining_pool',
                    'risk': None if results['is_mining_pool'] else 'medium'
                },
                {
                    'name': 'is_storing_phishing',
                    'risk': None if results['is_storing_phishing'] else 'high'
                },
                {
                    'name': 'is_phishing',
                    'risk': None if results['is_phishing'] else 'high'
                }
            )
        if type.lower() == 'ip':
            vulnerabilities.add(
                {
                    'name': 'is_cdn',
                    'risk': None if results['is_cdn'] else 'info'
                },
                {
                    'name': 'is_cnc',
                    'risk': None if results['is_cnc'] else 'critical'
                },
                {
                    'name': 'is_distributing_malware',
                    'risk': None if results['is_distributing_malware'] else 'critical'
                },
                {
                    'name': 'is_hosting',
                    'risk': None if results['is_hosting'] else 'info'
                },
                {
                    'name': 'is_known_attacker',
                    'risk': None if results['is_known_attacker'] else 'high'
                },
                {
                    'name': 'is_known_scanner',
                    'risk': None if results['is_known_scanner'] else 'medium'
                },
                {
                    'name': 'is_mining_pool',
                    'risk': None if results['is_mining_pool'] else 'medium'
                },
                {
                    'name': 'is_scanner',
                    'risk': None if results['is_scanner'] else 'medium'
                },
                {
                    'name': 'is_open_proxy',
                    'risk': None if results['is_open_proxy'] else 'medium'
                },
                {
                    'name': 'is_sinkhole',
                    'risk': None if results['is_sinkhole'] else 'low'
                },
                {
                    'name': 'is_tor_node',
                    'risk': None if results['is_tor_node'] else 'medium'
                },
                {
                    'name': 'is_vpn_node',
                    'risk': None if results['is_vpn_node'] else 'medium'
                },
            )

    else:
        errors = [
            f'There is an error in getting the data from Maltiverse',
            '',
            f'{results["message"]}',
            '',
            ''
        ]
        error_printer(True, errors)

    # return gathered data
    return [
        blocked_list,
        vulnerabilities
    ]



#!/usr/bin/env python


"""
    ### Subdomain: Hacker Target API

    This function gets the list of subdomains and corresponding IPs by calling
    the Hacker Target API.

    # Input:  - a single domain name
    # Output: - a dictionary contains SSL certificate details
"""


from colorama import Fore, Style

from config import config
from modules.utils import url_opener, printer
from modules.subdomain.utils import alt_name_sanitizer


def hacker_target(domain):
    # variables to store subdomains and related domains
    subdomains = set()
    related_domains = set()

    # print the subtitle: Hacker Target
    printer('      │        ├□ ' + Fore.GREEN +
            'Hacker Target API is calling' + Style.RESET_ALL)

    # download the result page of Hacker Target
    url = config['api']['hacker_target']['url_hosts'].format(domain)
    results = url_opener('GET', url, '', '', '', 'text', 'Hacker Target API')[0]
    results = results.split('\n')

    # get subdomains and related domains out from the output of the API call
    if results:
        subdomains, related_domains = alt_name_sanitizer(results, domain)

    # return subdomain and related domain
    return [
        subdomains,
        related_domains
    ]



#!/usr/bin/env python


"""
    ### General: site_analytics_parser

    This function looks for any site analytics tracking IDs within the HTML page.

    # Input:  - the HTML in the format of text
    # Output: - a dictionary contains all site analytics tracking IDs
"""


import re


def site_analytics_parser(html):
    # a variable to store site analytics tracking IDs
    analytics = dict()

    # parse analytics
    value = ''.join(re.findall('[\s"\']+(UA-[\d\-]*)["|\'\s]+', html, re.IGNORECASE))
    if value:
        analytics['Google'] = value

    # return results
    return analytics



#!/usr/bin/env python

from typing import List
from bs4 import BeautifulSoup


def nonce_finder(soup: BeautifulSoup) -> List[str]:
    """
    This function looks for any HTML <script> or <style> element has a nonce.
    CSP HTTP Header would be like: Content-Security-Policy: script-src 'nonce-{RANDOM}';
    HTML <script> element would be like: <script nonce="nonce-{RANDOM}"> CODE </script>

    Args:
        soup (BeautifulSoup): the BeautifulSoup format of the HTML page

    Returns:
        nonce (list): a list contains all unique nonces
    """
    nonces = set()

    # get all <style> and <script> elements that contain the 'nonce' attribute
    elements = soup.findAll('style', nonce=True)
    elements += soup.findAll('script', nonce=True)

    # iterate over elements
    for element in elements:
        # add the nonce to the set
        nonces.add(element['nonce'])

    # return results
    return list(nonces)



#!/usr/bin/env python

from colorama import Back, Fore, Style
# import csv
import json
import yaml
from typing import Dict

from config import config
import global_variable
from modules.utilities.printer import printer
from modules.utilities.error_printer import error_printer


def output_writer(file_name: str, formats: str, dict_data: Dict) -> None:
    """
    This function converts the given dictionary to different formats.
    If the format is TXT, it gets all STDOUT terms from 'ALL_PRINTS' then write into a file.
    If the format is JSON, it converts the dictionary of the data into a JSON data and then write into a file.
    If the format is YAML, it converts the dictionary of the data into a YAML data and then write into a file.
    If the format is CSV, it calls another function.

    Args:
        file_name (str): The name of the file to write the JSON data to.
        dict_data (dict): The dictionary containing the data to write to the file.
    """
    # Print the name of the domain
    printer(f'\n [{Fore.GREEN}☵{Fore.WHITE}]──{Fore.RED}{Back.WHITE} Writing into files.{Style.RESET_ALL}')

    # Iterate over formats
    i = 1
    for format in formats:
        # Form the extension of the file based on the selected "formant"
        if format in ('json_beautiful', 'json_b', 'b_json', 'beautiful_json'):
            extension = 'json'
        else:
            extension = format

        # Print the output location
        line = '└' if i == len(formats) else '├'
        printer(f'      {line}─── {Fore.CYAN}{file_name}.{extension}{Style.RESET_ALL}')

        # Write the result in the output file based on the selected format
        try:
            # Open the file and start writing in it
            with open(f"{file_name}.{extension}", mode='w', encoding='UTF8', newline='') as output_file:
                if format == "csv":
                    # output = csv.writer(output_file, quoting=csv.QUOTE_MINIMAL)
                    # start the recursive flattening process
                    # flatten_json(json_data)
                    continue

                # If the format is JSON (compact JSON)
                elif format == 'json':
                    output = json.dumps(dict_data, separators=(",", ":"), default=str, ensure_ascii=False)

                # If the format is JSON (Beautiful JSON)
                elif format in ['json_beautiful', 'json_b', 'b_json', 'beautiful_json']:
                    indent = config['output']['json']['indent']
                    output = json.dumps(dict_data, indent=indent, default=str, ensure_ascii=False)

                # If the format is YAML
                elif format == 'yaml':
                    indent = config['output']['yaml']['indent']
                    output = yaml.dump(dict_data, indent=indent, allow_unicode=True)

                # If the format is TXT (text)
                elif format == "txt":
                    output = global_variable.ALL_PRINTS

                # Print the output
                output_file.write(str(output))

        except IOError:
            texts = ['Error in in I/O (input/output).']
            error_printer('exception', texts)
        except TypeError:
            texts = [f'Error in the in serializing the "{format}" variable.']
            error_printer('exception', texts)
        except ValueError:
            texts = [f'Error in values are going to be written in the "{format}" file.']
            error_printer('exception', texts)
        except Exception:
            texts = [f'Error in writing into the "{format}" file.']
            error_printer('exception', texts)

        # Increment the counter of the formats
        i += 1

    # Print an empty line
    printer('\n')



#!/usr/bin/env python

from bs4 import BeautifulSoup
from datetime import datetime

from config import config
from modules.subdomain.utils import sub_related_domains
from modules.utilities.error_printer import error_printer
from modules.utilities.url_opener import url_opener


def dns_dumpster(domain):
    """
        ### DNS Checker: DNS Dumpster

        This function gets the value of DNS records provided by DNS Dumpster including
        MX, NS, TXT records as well as subdomain.

        # Input:  - a single domain name
        # Output: - a set of dictionaries contains a single DNS record details
                - a set contains list of subdomains
                - a set contains list of related domains
    """
    # some variables to store data
    dns_records = set()
    alt_names = set()
    subdomains = set()
    related_domains = set()

    # get the API url from the config file
    url = config['api']['dns_dumpster']['url']

    # get the CSRF from the page
    result = url_opener('GET', url, '', '', '', 'text', 'DNS Dumpster')[0]
    soup = BeautifulSoup(result.content, 'html.parser')
    csrf = soup.findAll('input', attrs={'name': 'csrfmiddlewaretoken'})[0]['value']

    # set a POST request to get the results
    cookies = {'csrftoken': csrf}
    headers = {'Referer': url}
    data = {'csrfmiddlewaretoken': csrf, 'targetip': domain, 'user': 'free'}
    post_request = url_opener('POST', url, cookies, data, headers, 'text', 'DNS Dumpster')

    # go further only if the POST request was successful
    if post_request.status_code == 200:
        # get all tables from the page
        tables = BeautifulSoup(post_request.content, 'html.parser').findAll('table')

        # iterate over NS rows
        for row in tables[0].findAll('tr'):
            columns = row.findAll('td')
            dns_records.add(
                {
                    'record': 'NS',
                    'value': str(columns[0]).split('<br/>')[0].split('>')[1].split('<')[0]
                }
            )
        # iterate over MX rows
        for row in tables[1].findAll('tr'):
            columns = row.findAll('td')
            dns_records.add(
                {
                    'record': 'MX',
                    'value': str(columns[0]).split('<br/>')[0].split('>')[1].split('<')[0]
                }
            )
        # iterate over A rows
        for row in tables[3].findAll('tr'):
            columns = row.findAll('td')
            dns_records.add(
                {
                    'record': 'A',
                    'value': str(columns[1]).split('<br/>')[0].split('>')[1].split('<')[0]
                }
            )
            alt_names.add(
                str(columns[0]).split('<br/>')[0].split('>')[1].split('<')[0]
            )
        # iterate over TXT rows
        for column in tables[2].findAll('td'):
            dns_records.add(
                {
                    'record': 'TXT',
                    'value': column.text
                }
            )

        # form the domain info list
        domain_info = [
            domain,
            'DNS Dumpster',
            datetime.now(),
            ''
        ]
        # call the function to define if the alt name is a subdomain or not
        subdomains, related_domains = sub_related_domains(alt_names, domain_info)

    # if there was an error with getting info
    else:
        texts = [
            'Open the DSN Dumpster faced an issue.',
            '',
            f'The error code is: {post_request.status_code}'
        ]
        error_printer(True, texts)

    # return gathered data
    return [
        dns_records,
        subdomains,
        related_domains
    ]



#!/usr/bin/env python


"""
    ### DNS: Utilities

    Here is the list of general functions used for DNS queries
"""


from config import config

import re


# get the rr_record and the answer, return the DNS record type
# i.e. for a TXT record, it might return SPF, DMARC or DKIM
def dns_record_type(rr_record, key, answer):
    # iterate over the list of terms used in TXT records
    # it is used to fill the 'type' filed
    if rr_record == 16:
        for txt_term in sorted(set(config['dns']['include_txt_records'])):
            if txt_term.lower() in [answer.lower(), key.lower()]:
                return txt_term.split(':')[0]
            else:
                return ''


# convert the DNS record to RR Type
def rr_record_finder(record):
    # check if record itself is an RR type
    if isinstance(record, int):
        rr_record = record
    else:
        rr_record = config['dns']['rr_dns_records'][record]
    
    return rr_record



#!/usr/bin/env python


"""
    ### Blocklist Checker

    This function 
    
    # Input:  - a single bacon
              - type of the bacon [ipv4, ipv6, domain]
    # Output: - a list of dictionary contains 
"""


from modules.whois.ip_api import ip_api
from modules.whois.utils import validate_ip
from modules.blocklist.dnsbl import dnsbl
from modules.blocklist.gray_noise import gray_noise


def blocklist_checker(bacon, type):

    type = type.lower()

    # return the results in the format of set
    return 



#!/usr/bin/env python

from colorama import Fore, Style
from config import config

from modules.utilities.printer import printer
from modules.utilities.exception_handling import exception_details


def error_printer(flag, texts):
    """
    This function prints errors related to the exceptions, based on verbosity

    Args:
        flag (str | boolean): "exception" or True to print based on verbosity, and
                              False to print nothing as the error
        texts (list): a list contains text for each verbosity in the following format
                      ["text for verbosity 1", ..., "text for verbosity 5"]
    """
    # If flag is True/exception and verbosity is not 0 (or False)
    if flag and config['verbosity']:
        # Print error messages for verbosity 1 (-v)
        if (
            config['verbosity'] >= 1
            and len(texts) >= 1
            and texts[0]
        ):
            verbosity = '      │        ├─1─■ '
            printer(f'{verbosity}{Fore.RED}{texts[0]}{Style.RESET_ALL}')

        # Print error messages for verbosity 2 (-vv)
        if (
            config['verbosity'] >= 2
            and len(texts) >= 2
            and texts[1]
        ):
            verbosity = '      │        ├─2─■ '
            printer(f'{verbosity}{Fore.RED}{texts[1]}{Style.RESET_ALL}')

        # Print error messages for verbosity 3 (-vvv)
        if (
            config['verbosity'] >= 3
            and len(texts) >= 3
            and texts[2]
        ):
            verbosity = '      │        ├─3─■ '
            printer(f'{verbosity}{Fore.RED}{texts[2]}{Style.RESET_ALL}')

        # Print error messages for verbosity 4 (-vvvv)
        if (
            config['verbosity'] >= 4
            and len(texts) >= 4
            and texts[3]
        ):
            verbosity = '      │        ├─4─■ '
            printer(f'{verbosity}{Fore.RED}{texts[3]}{Style.RESET_ALL}')

        # Print error messages for verbosity 5 (-vvvvv)
        if config['verbosity'] >= 5:
            if (
                len(texts) >= 5
                and texts[4]
            ):
                verbosity = '      │        ├─5─■ '
                printer(f'{verbosity}{Fore.RED}{texts[4]}{Style.RESET_ALL}')

            # In case, it is an exception error message
            if flag == 'exception':
                ex = exception_details()
                verbosity = '      │        ├─5─■ '
                printer(f'{verbosity}Exception Message: {Fore.MAGENTA}{ex[0]} ➜ '
                        f'{ex[4]}:{ex[3]}{Fore.RED} == {ex[2]} ⚊ {ex[1]}{Style.RESET_ALL}')



#!/usr/bin/env python


"""
    ### Whois: WhoisXML API

    This function gets the whois of the given domain.

    # Input:  - a single domain name
    # Output: - a dictionary contains robots.txt details
              - a list of name servers (NS)
"""

from datetime import datetime
from colorama import Fore, Style

from config import config
from modules.utils import url_opener, printer, date_formatter, json_key_checker


def whois_xml(domain):
    # a variable to store results
    domain_whois = dict()
    name_servers = list()

    # check if the whois API key is set
    if config['api']['WhoisXML']['api_key'] == '':
        printer('      │      ■ ' + Fore.RED + 'WhoisXML API key is not defined. Do it in the "config.py" file.' + Style.RESET_ALL)
        return domain_whois
    else:
        # get the date format of the WhoisXML from the config file
        date_format = config['api']['whois_xml']['date_format']

        # call the API
        url = config['api']['whois_xml']['url_whois'].format(config['api']['WhoisXML']['api_key'], domain)
        results = url_opener('GET', url, '', '', '', 'json', 'WhoisXML API')[0]['WhoisRecord']

        # convert string dates to the object date and format it accordingly
        # only if there is any data
        if results:
            domain_whois['validity']['create_date'] = date_formatter(json_key_checker(results, ['create_date', '']), date_format)
            domain_whois['validity']['update_date'] = date_formatter(json_key_checker(results, ['update_date', '']), date_format)
            domain_whois['validity']['expiration_date'] = date_formatter(json_key_checker(results, ['expiry_date', '']), date_format)
            domain_whois['validity']['age'] = (datetime.strptime(domain_whois['validity']['expiration_date'], config['date_format']) - datetime.strptime(domain_whois['validity']['create_date'], config['date_format'])).days
            domain_whois['validity']['past_days'] = (datetime.now() - datetime.strptime(domain_whois['validity']['create_date'], config['date_format'])).days
            domain_whois['validity']['left_days'] = (datetime.strptime(domain_whois['validity']['expiration_date'], config['date_format']) - datetime.now()).days

            domain_whois['registrar']['name'] = json_key_checker(results, ['registrarName', ''])
            domain_whois['registrar']['iana_id'] = json_key_checker(results, ['registrarIANAID', ''])
            domain_whois['registrar']['whois_server'] = json_key_checker(results, ['whoisServer', ''])

            domain_whois['registrant']['company'] = json_key_checker(results, ['registrant', 'organization'])
            domain_whois['registrant']['country'] = json_key_checker(results, ['registrant', 'country'])
            domain_whois['registrant']['state'] = json_key_checker(results, ['registrant', 'state'])

            domain_whois['administrative']['company'] = json_key_checker(results, ['administrativeContact', 'organization'])
            domain_whois['administrative']['country'] = json_key_checker(results, ['administrativeContact', 'country'])
            domain_whois['administrative']['state'] = json_key_checker(results, ['administrativeContact', 'state'])

            domain_whois['technical']['company'] = json_key_checker(results, ['technicalContact', 'organization'])
            domain_whois['technical']['country'] = json_key_checker(results, ['technicalContact', 'country'])
            domain_whois['technical']['state'] = json_key_checker(results, ['technicalContact', 'state'])

            domain_whois['billing']['company'] = json_key_checker(results, ['billing_contact', 'organization'])
            domain_whois['billing']['country'] = json_key_checker(results, ['billing_contact', 'country'])
            domain_whois['billing']['state'] = json_key_checker(results, ['billing_contact', 'state'])

            name_servers = json_key_checker(results, ['nameServers', 'hostNames'])

        return [
            domain_whois,
            name_servers
        ]



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Phish Tank

    This function checks if the domain is listed by Phish Tank or not

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains results if listed
"""


from config import config
from modules.utils import url_opener, error_printer


def phish_tank(domain):
    # variables to store results
    blocked_list = set()

    url = config['api']['phish_tank']['url']

    # different basic formats to send to Phish Tank
    # it is needed as Phish Tank doesn't support the wildcard search
    formats = [
        'http://' + domain, 'https://' + domain,
        'http://' + domain + '/', 'https://' + domain + '/'
    ]

    # iterate over different formats
    for uri in formats:
        data = {'url': uri, 'format': 'json'}
        results = url_opener(
            'POST', url, '', data, '', 'text', 'Phish Tank'
        )[0]

        # if there is no issue
        if results['meta']['status'] == 'success':
            # if the URL is listed in Phish Tank
            if results['results']['in_database']:
                blocked_list.add(
                    {
                        'by': 'Phish Tank',
                        'reason': 'Phishing',
                        'type': 'Domain',
                        'name': domain
                    }
                )
        # if there was an error with getting info
        else:
            texts = [
                f'There is an issue in getting data from Phish Tank.',
                '',
                f'The error is: {results["errortext"]}'
            ]
            error_printer(True, texts)

# return gathered data
    return blocked_list



#!/usr/bin/env python


"""
    ### IP Checker

    This function gets the IP whois, the IP version, and the list of sources
    blocked the IP.
    
    # Input:  a single IP address
    # Output: a dictionary contains IP address details
"""


from modules.whois.ip_api import ip_api
from modules.whois.utils import validate_ip
from modules.vulnerability.censys import Censys
from modules.blocklist.gray_noise import gray_noise
from blocklist_checker import blocklist_checker


def ip_checker(ip):
    # a variable to store results
    ip_details = dict()

    # call different APIs
    whois, tags, reverse_dns = ip_api(ip)
    censys = Censys(ip, 'ip')

    # get the IP version
    ip_details['ip'] = ip
    ip_details['version'] = f'IPv{validate_ip(ip)}'

    # get the whois
    ip_details['whois'] = whois

    # get the list of sources mentioned the IP as an anomaly from different sources
    ip_details['listed'] = blocklist_checker(ip, "ip")
    ip_details['listed'].update(gray_noise(ip))

    # get technologies
    ip_details['technologies'] = censys['technologies']

    # get open ports
    ip_details['ports'] = censys['ports']

    # get the reverse DNS
    if reverse_dns != censys['reverse_dns']:
        reverse_dns = [reverse_dns, censys['reverse_dns']]
    ip_details['reverse_dns'] = reverse_dns

    # get the tags
    ip_details['tags'] = tags

    # return the results in the format of set
    return ip_details



#!/usr/bin/env python

from colorama import Fore, Back, Style

from modules.ssl.openssl import openssl
from modules.ssl.crt_sh import crt_sh
from modules.ssl.ssl_labs import ssl_labs
from modules.utilities.printer import printer
from modules.utils import json_value


def ssl_checker(domain):
    """
        This function is responsible for getting SSL certificate details through
        different methods. Therefore, it will print the summary results on STDOUT.

        # Input:  a single domain name
        # Output: a dictionary contains SSL certificate details
    """
    # a variable to store cert info
    cert_info = set()

    # get the data and update the set
    cert_info = openssl(domain)
    cert_info.update(crt_sh(domain))
    cert_info.update(ssl_labs(domain))

    # print out found SSL info
    printer('      │\n      ├───' + Fore.BLACK + Back.WHITE + ' SSL Certificate Details ' + Style.RESET_ALL)

    # print the results
    if cert_info:
        printer('      │      ■ Issue Date:            ' + Fore.YELLOW + json_value(cert_info['validity'], 'issue_date') + Style.RESET_ALL)
        printer('      │      ■ Expiration Date:       ' + Fore.YELLOW + json_value(cert_info['validity'], 'expiration_date') + Style.RESET_ALL)
        printer('      │      ■ Validity:              ' + Fore.YELLOW + json_value(cert_info['validity'], 'past_days') + ' days past & ' + json_value(cert_info['validity'], 'left_days') + ' days remained' + Style.RESET_ALL)
        printer('      │      ■ Signature:             ' + Fore.YELLOW + json_value(cert_info, 'signature') + Style.RESET_ALL)
        printer('      │      ■ Fingerprint (SHA-256): ' + Fore.YELLOW + json_value(cert_info['fingerprint'], 'sha256') + Style.RESET_ALL)
        printer('      │      ■ Serial Number:         ' + Fore.YELLOW + json_value(cert_info, 'serial_number') + Style.RESET_ALL)
        printer('      │      ' + Fore.CYAN + '■ Issuer' + Style.RESET_ALL)
        printer('      └┐      ■■  Name:           ' + Fore.YELLOW + json_value(cert_info['issuer'], 'common_name') + Style.RESET_ALL)
        printer('       │      ■■  Org. Name:      ' + Fore.YELLOW + json_value(cert_info['issuer'], 'organization_name') + Style.RESET_ALL)
        printer('       │      ■■  Org. Unit Name: ' + Fore.YELLOW + json_value(cert_info['issuer'], 'organization_unit_name') + Style.RESET_ALL)
        printer('       │      ■■  Country:        ' + Fore.YELLOW + json_value(cert_info['issuer'], 'country') + Style.RESET_ALL)
        printer('       │      ■■  State:          ' + Fore.YELLOW + json_value(cert_info['issuer'], 'state') + Style.RESET_ALL)
        printer('       │      ■■  City:           ' + Fore.YELLOW + json_value(cert_info['issuer'], 'city') + Style.RESET_ALL)
        printer('      ┌┘      ■■  email:          ' + Fore.YELLOW + json_value(cert_info['issuer'], 'email_address') + Style.RESET_ALL)
        printer('      │      ' + Fore.CYAN + '■ Subject' + Style.RESET_ALL)
        printer('      └┐      ■■  Name:           ' + Fore.YELLOW + json_value(cert_info['subject'], 'common_name') + Style.RESET_ALL)
        printer('       │      ■■  Org. Name:      ' + Fore.YELLOW + json_value(cert_info['subject'], 'organization_name') + Style.RESET_ALL)
        printer('       │      ■■  Org. Unit Name: ' + Fore.YELLOW + json_value(cert_info['subject'], 'organization_unit_name') + Style.RESET_ALL)
        printer('       │      ■■  Country:        ' + Fore.YELLOW + json_value(cert_info['subject'], 'country') + Style.RESET_ALL)
        printer('       │      ■■  State:          ' + Fore.YELLOW + json_value(cert_info['subject'], 'state') + Style.RESET_ALL)
        printer('       │      ■■  City:           ' + Fore.YELLOW + json_value(cert_info['subject'], 'city') + Style.RESET_ALL)
        printer('      ┌┘      ■■  email:          ' + Fore.YELLOW + json_value(cert_info['subject'], 'email_address') + Style.RESET_ALL)
    else:
        printer('      │      ■■ ' + Fore.RED + 'No SSL certificate detail is found.' + Style.RESET_ALL)

    # return the result in the format of list instead of set
    return cert_info



#!/usr/bin/env python

from bs4 import BeautifulSoup

from config import config
from modules.utils import url_opener, date_formatter


def rapid_dns(bacon, type):
    """
        ### Vulnerability Scanner: Rapid DNS

        This function gets the list of historical DNS records and the reverse DNS of
        the given IP address or domain name.
        It contains duplicated subdomains and related domains as the date or the
        reason might be different.

        # Input:  - a single domain name or IP address
        # Output: - a set of dictionaries contains technologies used in the bacon
    """
    # variables to store results
    subdomains = set()
    related_domains = set()
    dns_records = set()

    # get the date format of the Rapid DNS from the config file
    date_format = config['api']['rapid_dns']['date_format']

    # get the results
    url = config['api']['rapid_dns']['url'].format(bacon)
    results = url_opener('GET', url, '', '', '', 'text', 'Rapid DNS')

    # find the table contains results using Beautiful Soup
    results = BeautifulSoup(results, 'html.parser')
    results = results.find(
        "table", {"class": "table table-striped table-bordered"}
    )
    results = results.find_all('tr')

    # parse the table and get the DNS data out of it
    for row in results:
        # add the found technology into the set
        row = row.find_all('td')

        if type.lower() == 'ip':
            related_domains.add(
                {
                    'value': row[0].text.strip(),
                    # why it is a related domain
                    'reason': row[2].text.strip(),
                    # the date that the relationship is found
                    'date': date_formatter(row[3].text.strip(), date_format)
                }
            )
        else:
            subdomains.add(
                {
                    'value': row[2].text.strip().split('.' + bacon)[0],
                    'reason': row[2].text.strip(),
                    'date': date_formatter(row[3].text.strip(), date_format)
                }
            )

    # return gathered data
    return [
        dns_records
    ]



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Full Hunt API

    This function checks if there is any information about the given domain name,
    including DNS records, open ports, ect.

    Read more: https://api-docs.fullhunt.io/

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains blocked lists
"""


from config import config
from modules.utils import url_opener, error_printer


def full_hunt(domain):
    # variables to store results
    port_status = set()
    vulnerabilities = set()

    # the HTTP headers to be sent to Full Hunt
    headers = {'X-API-KEY': config['api']['full_hunt']['api_key']}

    # form the URL based on the type of the domain
    url = config['api']['full_hunt']['url_domain'].format(domain)

    # get the results in JSON from Full Hunt
    results = url_opener('GET', url, '', '', headers, 'json', 'Full Hunt API')[0]

    # if the API call returns data
    if not 'status' in results:
        # get the blocked lists
        results = results['data']['hosts']

        # add the port to the set
        for port in results['network_ports']:
            port_status.add(
                {
                    'port': port,
                    'status': 'Open',
                    'protocol': '',
                    'service': '',
                }
            )
        vulnerabilities.add(
            {
                'name': 'is_cdn',
                'risk': None if results['is_cdn'] else 'info'
            },
            {
                'name': 'is_cloud',
                'risk': None if results['is_cloud'] else 'info'
            },
            {
                'name': 'is_cloudflare',
                'risk': None if results['is_cloudflare'] else 'info'
            },
            {
                'name': 'is_resolvable',
                'risk': None if results['is_resolvable'] else 'info'
            }
        )

    else:
        texts = [
            f'There was an error in getting data from Full Hunt',
            '',
            f'Code: {results["status"]} ➜ {results["message"]}'
        ]
        error_printer(True, texts)

    # return gathered data
    return [
        vulnerabilities,
        port_status
    ]



#!/usr/bin/env python

from bs4 import BeautifulSoup
from colorama import Fore, Back, Style
from typing import Dict

from config import config
from modules.general.html_meta import meta_parser
from modules.general.html_form import form_finder
from modules.general.html_links import link_finder
from modules.general.site_analytics import site_analytics_parser
from modules.http_header.csp import csp_parser
from modules.utilities.printer import printer
from modules.utilities.error_printer import error_printer
from modules.utilities.url_opener import url_opener
from modules.utilities.url_sanitizer import url_sanitizer


def site_status(domain: str) -> Dict[str, any]:
    """
    Retrieves general information and metadata about a website.

    Args:
        domain: The domain of the website to retrieve information from.

    Returns:
        A dictionary containing information and metadata about the website.
    """
    # A variable to store HTML data
    html_data = {
        'meta': {},
        # CSP headers are added as they might not be used in the page
        # However, we need to parse them later
        'http_headers': [],
        'redirects': [],
        'analytics': {},
        'csp_nonces': [],
        'csp_hashes': [],
    }

    # Print the title of the section
    printer(f'      ├───{Fore.BLACK}{Back.WHITE} General Info {Style.RESET_ALL}')

    # Sanitize the URL
    url = url_sanitizer(domain)[0]

    # Open the page
    text_data, status_code, history, headers, version, url = url_opener(
        'GET', url, '', '', '', 'text', 'General Site Status function',
    )

    # Continues only if there is no error; HTTP status code 2xx and 3xx
    if status_code < 200 or status_code >= 400:
        texts = [
            'Could not parse the HTML page.',
            f'HTTP Status Code: {status_code}',
            f'History: {history}',
            f'HTTP Headers: {headers}',
            f'Request Result: {text_data}',
        ]
        error_printer(True, texts)
        html_data['status_code'] = int(status_code)

    # If there is no error
    else:
        # Parse the HTML metadata
        soup = BeautifulSoup(text_data, "html.parser")

        # Get redirects
        if config['scan_type']['http_switch'][0] and history:
            html_data['redirects'].append(domain)
            # iterate over the redirection history
            for step in history:
                html_data['redirects'].append(step.url)

        # Get site analytics
        if config['scan_type']['http_switch'][1]:
            html_data['analytics'] = site_analytics_parser(text_data)

        # Parse the meta data and add them to the dictionary
        if config['scan_type']['http_switch'][2]:
            html_data['meta'] = meta_parser(soup)

        # Find any links in the page
        if config['scan_type']['http_switch'][3]:
            html_data['links'] = link_finder(soup, domain)

        # Find any forms in the page
        if config['scan_type']['http_switch'][4]:
            html_data['forms'] = form_finder(soup)

        # Parse the 'title' element and other data and write into a dict variable
        html_data['title'] = soup.find('title').text.strip() if soup.find('title') else ''
        html_data['status_code'] = int(status_code)
        html_data['version'] = version

        # Print the result
        printer(f'      │      ■ Domain:                 {Fore.YELLOW}{domain}{Style.RESET_ALL}')
        printer(f'      │      ■ Scanned URL:            {Fore.YELLOW}{url}{Style.RESET_ALL}')
        if 'status_code' in html_data and html_data['status_code']:
            printer('      │      ■ HTTP Status:            '
                    f'{Fore.YELLOW}{html_data["status_code"]}{Style.RESET_ALL}')
        if 'redirects' in html_data and html_data['redirects']:
            printer('      │      ■ HTTP Redirects:         '
                    f'{Fore.YELLOW}{" ➜ ".join(html_data["redirects"])}{Style.RESET_ALL}')
        if 'title' in html_data and html_data['title']:
            printer('      │      ■ Site Title:             '
                    f'{Fore.YELLOW}{html_data["title"]}{Style.RESET_ALL}')
        if 'links' in html_data and html_data['links']:
            printer('      │      ■ No. of Links:           '
                    f'Internal -> {Fore.YELLOW} {html_data["links"]["count"]["internal"]}{Style.RESET_ALL}  |  '
                    f'External -> {Fore.YELLOW} {html_data["links"]["count"]["external"]}{Style.RESET_ALL}')

            # Print the number of HTML element
            if 'count' in html_data['links'] and html_data['links']['count']:
                printer(f'      │\n      ├───{Fore.BLACK}{Back.WHITE} HTML elements with Links {Style.RESET_ALL}')

                # Iterate over HTTP elements
                for element in html_data['links']['count']:
                    if (
                        not element.startswith(tuple(['int_', 'ext_', 'internal', 'external']))
                        and html_data["links"]["count"][element]
                    ):
                        printer(f'      │      ■ No. of Elements {"<" + element + ">:":11}'
                                f'{Fore.YELLOW} {html_data["links"]["count"][element]}{Style.RESET_ALL}')

        # Print metadata
        if 'meta' in html_data and html_data['meta']:
            printer(f'      │\n      ├───{Fore.BLACK}{Back.WHITE} HTML Metadata {Style.RESET_ALL}')

            # Iterate over metadata
            for meta in html_data['meta']:
                for key, value in meta.items():
                    key = key.replace('_', ' ').replace('-', ' ').title()
                    printer(f'      │      ■ {key + ":":24}{Fore.YELLOW}{value}{Style.RESET_ALL}')

        # Print analytics
        if 'analytics' in html_data and html_data['analytics']:
            printer(f'      │\n      ├───{Fore.BLACK}{Back.WHITE} Site Analytics Tracking IDs {Style.RESET_ALL}')

            # Iterate over the analytics codes
            for key, value in html_data['analytics'].items():
                key = key.replace('_', ' ').replace('-', ' ').title()
                printer(f'      │      ■ {key + ":":24}{Fore.YELLOW}{value}{Style.RESET_ALL}')

        # Print HTTP headers
        if config['scan_type']['http_switch'][5] and headers:
            printer(f'      │\n      ├───{Fore.BLACK}{Back.WHITE} HTTP Headers {Style.RESET_ALL}')

            # Add the HTTP header to the list and print it
            for key, value in headers.items():
                key = key.lower()
                csp_flag = True

                # Get the CSP
                if key in ('content-security-policy', 'content-security-policy-report-only'):
                    csp_flag = False
                    # Print the results
                    printer(f'      │      ■ {key}')
                    # Parse CSP as well as get nonces and hashes
                    csp_value, nonces, hashes = csp_parser(value, soup)
                    # Write the CSP results in the output
                    html_data['http_headers'].append(
                        {
                            'name': key,
                            'value': csp_value
                        }
                    )
                    # Add nonces and hashes to the output
                    html_data['csp_nonces'] += nonces
                    html_data['csp_hashes'] += hashes

                # Other HTTP headers rather than CSP or CSP-Report-Only
                else:
                    # Print the results
                    printer(f'      │      ■ {key + ":":24}{Fore.YELLOW}{value}{Style.RESET_ALL}')
                    # Write the results in the output
                    html_data['http_headers'].append(
                        {
                            'name': key,
                            'value': value
                        }
                    )
            # If CSP or CSP-Report-Only is not set
            if csp_flag:
                # Write the error message of no CSP is set
                html_data['http_headers'].append(
                    {
                        'name': '__WARNING__',
                        'value': 'NO_CSP'
                    }
                )
        # If there HTTP header is not retrieved or doesn't exist
        elif config['scan_type']['http_switch'][5] and not headers:
            html_data['http_headers'] = {
                'name': '__ERROR__',
                'value': 'NO_HTTP_HEADER'
            }

        # Get CSP headers in HTML meta element; it should contain the attribute 'http-equiv'
        for csp in ('content-security-policy', 'content-security-policy-report-only'):
            # Find <meta> elements contains the attribute 'http-equiv' with certain values
            csp_header = soup.find('meta', {'http-equiv': csp})
            if csp_header:
                # Parse CSP as well as get nonces and hashes
                csp_value, nonces, hashes = csp_parser(csp_header.attrs['content'], soup)
                # Write the results in the output
                html_data['http_headers'].append(
                    {
                        'name': csp,
                        'value': csp_value
                    }
                )
                # Add nonces and hashes to the output
                html_data['csp_nonces'] += nonces
                html_data['csp_hashes'] += hashes

        # Remove duplications from the list
        html_data['csp_nonces'] = list(set(html_data['csp_nonces']))
        html_data['csp_hashes'] = list(set(html_data['csp_hashes']))

    # Return results
    return [
        html_data,
        url
    ]



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Pulse Dive

    This function gets the scan results from Pulse Dive.

    # Input:  - a single domain name
    # Output: - a dictionary contains the domain
"""


from datetime import datetime
from time import sleep

from config import config
from modules.utils import url_opener, date_formatter
from modules.subdomain.utils import sub_related_domains


def pulse_dive(domain):
    # variables to store results
    vulnerabilities = set()
    port_status = set()
    technologies = set()
    http_headers = set()
    html_meta = set()
    alt_names = set()
    related_domains = set()
    subdomains = set()

    # get the scan type (active or passive) and the delay
    scan_type = 1 if config['api']['pulse_dive']['scan_type'].lower() == 'active' else 0
    delay = config['api']['pulse_dive']['delay']
    date_format = config['api']['pulse_dive']['date_format']

    # form the data to be sent to the API (start the scan)
    data = {
        "value": domain,
        "probe": scan_type,
        "pretty": "0",
        "key": config['api']['pulse_dive']['api_key']
    }

    # send the scan request to Pulse Dive and get the scan ID
    url = config['api']['pulse_dive']['url_scan']
    results = url_opener('POST', url, '', data, '', 'json', 'Pulse Dive')
    scan_id = results[0]['qid']

    url = config['api']['pulse_dive']['url_result'].format(
        scan_id,
        config['api']['pulse_dive']['api_key']
    )

    # check if result is ready
    while True:
        # send the HTTP request to scan
        results = url_opener('GET', url, '', '', '', 'json', 'Pulse Dive')[0]

        # if the doesn't exist, the report exits
        if results['status'] == 'done':
            break
        else:
            # sleep for a certain time to finish the scan
            sleep(delay)

    # get the results
    results = results['data']

    # get the list of vulnerabilities
    for risk_factor in results['riskfactors']:
        # define the risk level
        if risk_factor['riskfactors']['risk'] in ('none', 'unknown'):
            risk = 'info'
        else:
            risk_factor['riskfactors']['risk']
        # add the results to vulnerabilities
        vulnerabilities.add(
                {
                    'name': risk_factor['description'],
                    'risk': risk
                }
        )

    # get the list of open ports
    for port in results['attributes']['port']:
        port_status.add(
            {
                'port': port,
                'status': 'open',
                'protocol': '',
                'service': '',
            }
        )

    # get the list of technologies
    for tech in results['attributes']['technology']:
        technologies.add(
            {
                'name': tech,
                'version': '',
                'category': '',
                'date': date_formatter(results['stamp_updated'], date_format)
            }
        )

    # get the list of HTTP headers
    for name, value in results['properties']['http'].items():
        # ignore the illegal headers
        if name.startswith('++'):
            continue
        # add the HTTP header to the list
        http_headers.add(
            {
                'name': name.lower(),
                'value': value,
            }
        )

    # get the list of HTML meta data
    for name, value in results['properties']['meta'].items():
        html_meta.add(
            {
                'name': name.lower(),
                'value': value,
            }
        )

    # get list of the alternative names
    for alt_name in results['links']['Related Domains']:
        alt_names.add(
            {
                'name': alt_name['indicator'],
                'reason': 'Pulse Dive',
                'date': date_formatter(alt_name['stamp_linked'], date_format)
            }
        )

    # call the function to extract subdomain & related-domains from
    # the alternative names
    subdomains, related_domains = sub_related_domains(alt_names, domain)

    # return gathered data
    return [
        {
            "name": domain,
            "technologies" : technologies,
            "ports": port_status,
            "vulnerabilities": vulnerabilities,
            "http_headers": http_headers,
            "http_meta": html_meta,
            "subdomains": subdomains,
            "related_domains": related_domains
        }
    ]



#!/usr/bin/env python

import dns.resolver
import requests
import random
from config import config
from modules.utilities.url_sanitizer import url_sanitizer


# check if the subdomain is vulnerable to the takeover by misconfiguration in CNAME
# results schema:
#       0: no subdomain takeover vulnerability
#       1: subdomain takeover vulnerability
#       2: there was an error
def subdomain_takeover(rdata):
    # sanitize the rdata
    rdata = url_sanitizer(rdata)[1]

    # iterate over the CNAMEs defined in the config file
    for cname in config['cnames']:
        flag = True
        headers = {'User-Agent': random.choice(config['user_agents'])}
        f_type = config['cnames'][cname]['f_type'].lower()

        # if the found CNAME target is in our list
        if cname in rdata:
            try:
                # act based on the type of the fingerprint
                if f_type == 'content':
                    try:
                        req = requests.get('https://' + rdata, headers=headers).text
                    # in case the scheme is not HTTPS
                    except requests.exceptions.InvalidSchema:
                        req = requests.get('http://' + rdata, headers=headers).text
                        # return error ocurred
                        return 2
                    except Exception:
                        flag = False

                    # if the page is loaded and the fingerprint is found
                    if not flag:
                        fingerprint = config['cnames']['cname']['fingerprint'].lower()
                        if fingerprint in req:
                            # return potential subdomain takeover
                            return 1
                        else:
                            # return no subdomain takeover vulnerability
                            return 0
                # if the error is: the domain doesn't exist
                elif f_type == 'nxdomain':
                    try:
                        dns.resolver.resolve(rdata, 'A')
                        # return no subdomain takeover vulnerability
                        return 0
                    except dns.resolver.NXDOMAIN:
                        # return potential subdomain takeover
                        return 1
                    except Exception:
                        # return error ocurred
                        return 2
                # if it shows 404 HTTP error
                elif f_type == 'http_status':
                    try:
                        req = requests.get('https://' + rdata, headers=headers).text
                    # in case the scheme is not HTTPS
                    except requests.exceptions.InvalidSchema:
                        req = requests.get('http://' + rdata, headers=headers).text
                        # return error ocurred
                        return 2
                    except Exception:
                        # return error ocurred
                        return 2

                    # if the fingerprint is what is defined in the config file
                    fingerprint = config['cnames']['cname']['fingerprint']
                    if fingerprint == str(req.status_code):
                        # return potential subdomain takeover
                        return 1
                    else:
                        # return no subdomain takeover vulnerability
                        return 0
            # exceptions
            except Exception:
                # return error ocurred
                return 2



#!/usr/bin/env python


"""
    ### Subdomain: Utilities

    Here is the list of general functions used for finding subdomains
"""


from modules.utilities.error_printer import error_printer


# this function checks if the given alternative name is a subdomain or a related
# domain.
#
# domain should be a dictionary in this format
# domain_info = {'name':string, 'reason':string, 'date':string}
#
# result would be a list. The first items represent the type, and the second item
# represents the value.
# possible output
#    0: alt name and domain are the same
#    1: alt name is a subdomain
#    2: alt name is a related domain
def sub_related_domain(alt_name, domain):
    # strip the illegal chars from the alternative name
    alt_name = str(alt_name['name']).lower().strip()
    alt_name = alt_name.replace('https://', '').replace('http://', '').replace('/', '')

    if alt_name != domain.lower():
        # sanitize (subject) alternative names
        alt_name = alt_name.replace('*.', '')
        # for OpenSSL "subjectAltName"
        alt_name = alt_name.replace('DNS:', '')

        # check if it is an alt name or a related domain
        # then add it to the corresponding set
        if alt_name.endswith(domain):
            return [
                1,
                {
                    'name': alt_name.split('.' + domain)[0],
                    'reason': alt_name['reason'],
                    'date': alt_name['date']
                }
            ]
        # if it is a related domain
        else:
            return [
                2,
                {
                    'name': alt_name,
                    'reason': alt_name['reason'],
                    'date': alt_name['date']
                }
            ]
    # if the domain name and the alternative name are the same
    else:
        return [0, False]


# get the list of alternative domains and return the 2 sets
# one for subdomains and one for related domains
# domain_info should be a list in this format
# domain_info = [domain, reason, date, date_format]
def sub_related_domains(alt_names, domain_info):
    # variables to store subdomains and related domains
    subdomains = set()
    related_domains = set()

    # continue only if there is any data for it
    if alt_names:
        for alt_name in alt_names:
            # iterate over the found alternative names
            # call the function to define if the alt name is a subdomain or not
            alt_name_result = sub_related_domain(alt_name, domain_info)
            # if there is no results
            if alt_name_result[0] == 0:
                continue
            # if it is a subdomain
            elif alt_name_result[0] == 1:
                subdomains.add(alt_name_result[1])
            # if it is a related-domain
            elif alt_name_result[0] == 2:
                related_domains.add(alt_name_result[1])
            # if there was an error with getting info
            else:
                texts = [
                    f'There was an error in analyzing the alternative name',
                    '',
                    f'Code: {alt_name_result[0]} ➜ {alt_name_result[1]}'
                ]
                error_printer(True, texts)

    # return subdomain and related domain
    return [
        subdomains,
        related_domains
    ]



#!/usr/bin/env python

from typing import List

from config import config


def nonce_validator(source: str, nonces: List[str]) -> str:
    """
    This function validates a nonce of a given source and returns a term to show if it is matched
    with the value of the nonce attribute in the HTML element.

    Args:
        source (str): The source string containing the nonce.
        nonces (list): The list of nonces

    Returns:
        str: The string defined in config file that shows if the nonce is matched or not.
    """
    # check if the nonce is matched with the gathered nonces from HTML <script> or <style> elements
    if nonces:
        if source.replace('nonce-', '') in nonces:
            return config['html']['nonce']['matched']
        else:
            return config['html']['nonce']['unmatched']
    # if there is no element with the attribute 'nonce' while it is set in the CSP header as the source
    else:
        return config['html']['nonce']['unmatched_']



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Check Host

    This function gets the open ports

    Read more: - https://check-host.net/about/api

    # Input:  - a single domain name or IP address
    # Output: - a set contains open TCP/UDP ports
"""


from time import sleep

from config import config
from modules.utils import url_opener


def criminal_ip(bacon, protocol):
    # variables to store subdomains and related domains
    port_status = set()

    # the HTTP headers to send to Check Host
    headers = {'Accept': 'application/json'}

    # get the list of the ports
    if protocol == 'TCP':
        ports = config['scan_type']['tcp_ports']
    elif protocol == 'UDP':
        ports = config['scan_type']['udp_ports']
    else:
        ports = [21, 22, 25, 80, 143, 443, 587, 993, 995, 2082, 2077, 3306]

    # do scan for each port
    for port in ports:
        # add the port number to the bacon
        bacon = f'{bacon}:{str(port)}'
        # form the url to be requested
        if protocol == 'UDP':
            url = config['api']['check_host']['url_udp_port'].format(bacon)
        else:
            url = config['api']['check_host']['url_tcp_port'].format(bacon)
        # request the port scan from Check Host
        request_id = url_opener('GET', url, '', '', headers, 'json', False)
        request_id = request_id[0]['request_id']

        # sleep for a certain time to
        sleep(config['api']['check_host']['delay'])

        # get the results
        url = config['api']['check_host']['url_result'].format(request_id)
        results = url_opener(
            'GET', url, '', '', headers, 'json',
            f'Check Host API for port {port}'
        )

        # set the
        status = 'closed'

        # iterate over the result of each probe
        for result in results:
            # if any probes says that the port is open, see it as open
            if result and 'time' in result[0]:
                port_status = 'open'
                break

        # add the port to the set
        port_status.add(
            {
                'port': port,
                'status': status,
                'protocol': protocol,
                'service': '',
            }
        )

    # return gathered data
    return port_status



#!/usr/bin/env python

from colorama import Back, Fore, Style

from config import config
from modules.utilities.exception_handling import exception_details
from modules.utilities.printer import printer
from modules.utilities.printer import ALL_PRINTS


def txt_writer(file_name: str) -> None:
    """
    This function gets all STDOUT terms from 'ALL_PRINTS' then write into a file.

    Args:
        file_name (str): The name of the file to write the JSON data to.
    """
    try:
        # print the name of the domain
        printer(f'\n [{Fore.GREEN}+{Fore.WHITE}]──┬──{Fore.RED}{Back.WHITE} Write into '
                f'{file_name}{Style.RESET_ALL}{Fore.GREEN}{Style.RESET_ALL}\n      │')

        # open the file and start writing in it
        with open(file_name, mode='w', encoding='UTF8', newline='') as output_file:
            # write the line into the CSV file
            output_file.write(ALL_PRINTS)

    except IOError:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in in I/O (input/output).{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
                    f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
    except TypeError:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in the in serializing the TXT variable.{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
    except ValueError:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in values are going to be written in the TXT file.{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
    except Exception:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in writing into the TXT file.{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]}{Fore.RED}  {ex[1]}{Style.RESET_ALL}')
    finally:
        # print a message shows the scan for the domain is finished
        printer(f'      │\n      └──{Fore.RED}{Back.WHITE} Finish Writing in "{file_name}" {Style.RESET_ALL}\n')



#!/usr/bin/env python


"""
    ### DNS Records: DNS Blocklist Checker

    This function checks the domain name, IPv4, or IPv6 against different
    different sources (defined in the config.py file) to find if it is blocked
    by any blocklist maintainer or not.

    # Input:  - a list of inputs: domain names, IPs (version 4 or 6)
              - the type of the input
    # Output: - a set of values related to the given DNS record
"""


import ipaddress

from config import config
from modules.dns.dns_resolver import dns_resolver
from modules.utilities.url_sanitizer import url_sanitizer


def dnsbl(bacons, type):
    # a variable to store results
    blocked_list = set()
    # assure that bacons is the list
    bacons = list(bacons)

    # form bacons
    if not isinstance(bacons, list):
        bacons = [bacons]

    # iterate over bacons
    for bacon in bacons:
        if type.lower() in ['ipv4', 'ipv6']:
            bacon = ipaddress.ip_address(bacon).reverse_pointer
            # strip the additional texts
            bacon = bacon.replace('in-addr.arpa', '').replace('ip6.arpa', '')
        elif type in ['domain']:
            # sanitize the domain name
            bacon = url_sanitizer(bacon)[1]
            bacon = bacon + '.'

        for bl, content in config['dnsbl'].items():
            if type in content['type']:
                # get the result of the DNS query "A" without printing it
                answer = dns_resolver(bacon + bl, 'A')
                if answer:
                    blocked_list.add({
                        'by': bl,
                        'reason': content[answer[0]][1].split(':')[0],
                        'type': type,
                        'name': bacon
                    })

    # return results
    return blocked_list



#!/usr/bin/env python

import dns.resolver
from colorama import Fore, Style

from config import config
from modules.dns.dns_resolver import dns_resolver
from modules.utilities.printer import printer


# get and validate RRSIG as it can't be retrieved directly by querying 'RRSIG'
# it needs a UDP query on port 53
def dnssec(domain):
    # set the print arguments for the function "url_opener"
    print_args = [True, '      │        ├──■ ', '      │        │  ■■ ']
    # get the nameserver (NS) and A records of the domain
    ns_server = dns_resolver(domain, 'NS', print_args).rrset[0].to_text()
    a_record = dns_resolver(ns_server, 'A', print_args).rrset[0].to_text()
    # get the DNSKEY of the domain
    dnskey = dns.message.make_query(domain, 'DNSKEY', want_dnssec=True)

    # send a UDP query on port 53 to get the RRSIG
    response = dns.query.udp(dnskey, a_record)

    # if there is any error based on the DNS RCODEs
    if response.rcode() != 0:
        error = config['dns']['dns_rcode'][response.rcode()]
        printer('      │       ■ ' + Fore.RED +
                'There is an error: {0}.'.format(error) + Style.RESET_ALL)
    # parse the RRSIG for the DNSKEY from the response
    else:
        dns_sec = response.answer

        # if the array doesn't contain the DNSKEY and RRSIG
        if len(dns_sec) != 2:
            printer('      │       ■ ' + Fore.RED +
                    'Could not find any DNSKEY and RRSIG record in the zone' + Style.RESET_ALL)
        else:
            printer(dns_sec)
            printer('      │       ■■  ' + Fore.YELLOW + str(dns_sec[1]) + Style.RESET_ALL)

            # resolve the DNSSEC
            try:
                name = dns.name.from_text(domain)
                dns.dnssec.validate(dns_sec[0], dns_sec[1], {name:dns_sec[0]})
            except dns.dnssec.ValidationFailure:
                printer('      │' + ' ' * 14 + '└□ ' + Fore.RED +
                        'The DNSSEC signature (RRSIG) is invalid.' + Style.RESET_ALL)
            except dns.dnssec.UnsupportedAlgorithm:
                printer('      │' + ' ' * 14 + '└□ ' + Fore.CYAN +
                        'The DNSSEC algorithm is not supported.' + Style.RESET_ALL)
            else:
                printer('      │' + ' ' * 14 + '└□ ' + Fore.GREEN +
                        'The DNSSEC is valid.' + Style.RESET_ALL)

            return str(dns_sec[1])

    return ''



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Geek Flare

    This function checks if the given domain supports different versions of HTTP,
    including ver. 1.0, ver. 1.1, ver. 2.0, and ver. 3.0.

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains HTTP headers
"""


from config import config
from modules.utils import url_opener, error_printer


def geek_flare_broken_link(domain):
    # a variable to store results
    versions = dict()

    data = {
        "url": domain,
        "followRedirect": True
    }
    headers = {
        'x-api-key': config['api']['geek_flare']['api_key'],
        'Content-Type': 'application/json'
    }

    # check if report exist
    url = config['api']['geek_flare']['url_http_protocol']
    results = url_opener('POST', url, '', data, headers, 'json', 'Geek Flare')

    # check if there is any error
    if results['apiCode'] != 200:
        # add the found versions to the dictionary
        versions = results['data']

    else:
        errors = [
            f'There is an error in getting the data from Geek Flare',
            '',
            f'{results["message"]}',
            '',
            ''
        ]
        error_printer(True, errors)

    # return gathered data
    return versions



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Full Hunt API

    This function returns all the subdomains found by Full Hunt.

    Read more: https://api-docs.fullhunt.io/

    # Input:  - a single domain name
    # Output: - a set of dictionaries contains the alternative domains
"""


from datetime import datetime

from config import config
from modules.utils import url_opener, error_printer, date_formatter
from modules.subdomain.utils import sub_related_domains


def full_hunt(domain):
    # variables to store results
    alt_names = set()

    # the HTTP headers to be sent to Full Hunt
    headers = {'X-API-KEY': config['api']['full_hunt']['api_key']}

    # form the URL based on the type of the domain
    url = config['api']['full_hunt']['url_subdomain'].format(domain)

    # get the results in JSON from Full Hunt
    results = url_opener('GET', url, '', '', headers, 'json', 'Full Hunt API')[0]

    # if the API call returns data
    if results['status'] == 200:
        # get the date in epoch format
        date = ['data']['metadata']['last_scanned']
        date = datetime.fromtimestamp(date)

        # get list of the alternative names
        for alt_name in results['subdomains']:
            alt_names.add(
                {
                    'name': alt_name,
                    'reason': 'Full Hunt',
                    'date': date_formatter(date, '')
                }
            )

        # call the function to extract subdomain & related-domains from
        # the alternative names
        subdomains, related_domains = sub_related_domains(alt_names, domain)
    else:
        texts = [
            'There was an error in getting data from Full Hunt',
            '',
            f'Code: {results["status"]} ➜ {results["message"]}'
        ]
        error_printer(True, texts)

    # return gathered data
    return [
        subdomains,
        related_domains
    ]



#!/usr/bin/env python

from config import config
from modules.utilities.url_opener import url_opener
from modules.whois.utils import validate_ip


def ip_api(ip):
    """
        ### Whois: IP API

        This function gets the whois of the given IP as well as some checks
        like if the IP hosts sites, provides VPN, or belongs to a mobile device.

        # Input:  - a single IP address
        # Output: - a dictionary contains whois details
                - a set contains tags related to the IP such as if the IP is used
                    for hosting, VPN, etc.
                - a string returns the reversed DNS of the given IP
    """
    # a variable to store results
    ip_whois = dict()
    vulnerabilities = set()
    reverse_dns = ''

    # check if the ip is a valid one
    if not validate_ip(ip):
        return ip_whois

    # call the API
    url = config['api']['ip_api']['url_lookup'].format(ip, config['api']['ip_api']['fields'])
    results = url_opener('GET', url, '', '', '', 'json', 'IP API')[0]

    # write the data into the dictionary
    if results:
        ip_whois['name'] = ip
        ip_whois['continent'] = results['continent']
        ip_whois['continent_code'] = results['continentCode']
        ip_whois['country'] = results['country']
        ip_whois['country_code'] = results['countryCode']
        ip_whois['region'] = results['regionName']
        ip_whois['district'] = results['district']
        ip_whois['city'] = results['city']
        ip_whois['zip_code'] = results['zip']
        ip_whois['latitude'] = results['lat']
        ip_whois['longitude'] = results['lon']
        ip_whois['timezone'] = results['timezone']
        ip_whois['time_offset'] = results['offset']
        ip_whois['isp'] = results['isp']
        ip_whois['organization'] = results['org']
        ip_whois['as_number'] = results['as']
        ip_whois['as_name'] = results['asname']

        reverse_dns = results['reverse']

        vulnerabilities.add(
            {
                'name': 'is_mobile',
                'risk':  None if results['mobile'] else 'info'
            },
            {
                'name': 'is_hosting',
                'risk':  None if results['hosting'] else 'info'
            },
            {
                'name': 'is_proxy',
                'risk': None if results['proxy'] else 'medium'
            },
        )

    # return the whois of the IP
    return [
        ip_whois,
        vulnerabilities,
        reverse_dns
    ]



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Criminal IP API

    This function gets the list of subdomains and corresponding IPs by calling
    the Criminal IP API.

    Read more: - https://www.criminalip.io/developer/api/

    # Input:  - a single domain name or IP address
              - the type of the asset that needs to be assessed
    # Output: - a set contains vulnerability assessment of the given domain
              - a set contains technologies used by the given domain
              - a set contains subdomains
              - a set contains related domains
              - a dictionary contains some details about the IP
"""


from colorama import Fore, Style

from config import config
from modules.utils import url_opener, printer, error_printer
from modules.subdomain.utils import alt_name_sanitizer


def criminal_ip(bacon, type):
    # variables to store subdomains and related domains
    vulnerabilities = set()
    technologies = set()
    subdomains = set()
    related_domains = set()
    ip_data = dict()
    ip_data['vulnerabilities'] = set()
    ip_data['ports'] = set()

    # print the subtitle: Criminal IP
    printer('      │        ├□ ' + Fore.GREEN + 'Criminal IP API is calling' + Style.RESET_ALL)

    # for the HTTP headers to send to Criminal IP
    headers = {'x-api-key': config['api']['criminal_ip']['api_key']}

    # get the results in JSON from Criminal IP
    if type == 'domain':
        url = config['api']['criminal_ip']['url_domain'].format(bacon)
        results = url_opener('GET', url, '', '', headers, 'json', 'Criminal IP API')[0]

        # if the API call returns data
        if results['status'] == 200:
            # get the latest report ID
            report_id = results['data']['reports'][0]['scan_id']

            url = config['api']['criminal_ip']['url_domain_report'].format(report_id)
            report = url_opener('GET', url, '', '', headers, 'json', 'Criminal IP API')[0]

            # if the API call returns data
            if report['status'] == 200:
                # get the list of subdomains and related domains
                for subdomain in report['data']['subdomains']:
                    subdomains.add(subdomain['subdomain_name'])

                subdomains, related_domains = alt_name_sanitizer(subdomains, bacon)

                # get the vulnerability data
                summary = summary
                vulnerabilities.add(
                    {
                        'name': 'connect_to_ip_directly',
                        'risk': None if summary['connect_to_ip_directly'] else 'info'
                    },
                    {
                        'name': 'diff_domain_favicon',
                        'risk': None if summary['diff_domain_favicon'] else 'low'
                    },
                    {
                        'name': 'double_slash_url',
                        'risk': None if summary['double_slash_url'] else 'info'
                    },
                    {
                        'name': 'fake_domain',
                        'risk': None if summary['fake_domain'] else 'medium'
                    },
                    {
                        'name': 'fake_https_url',
                        'risk': None if summary['fake_https_url'] else 'high'
                    },
                    {
                        'name': 'fake_ssl',
                        'risk': None if summary['fake_ssl']['invalid'] else 'high'
                    },
                    {
                        'name': 'hidden_element',
                        'risk': None if summary['hidden_element'] else 'low'
                    },
                    {
                        'name': 'hidden_iframe',
                        'risk': None if summary['hidden_iframe'] else 'medium'
                    },
                    {
                        'name': 'js_obfuscated',
                        'risk': None if summary['js_obfuscated'] else 'low'
                    },
                    {
                        'name': 'mail_server',
                        'risk': None if summary['mail_server'] else 'info'
                    },
                    {
                        'name': 'many_subdomain',
                        'risk': None if summary['many_subdomain'] else 'low'
                    },
                    {
                        'name': 'mitm_attack',
                        'risk': None if summary['mitm_attack'] else 'critical'
                    },
                    {
                        'name': 'newborn_domain',
                        'risk': None if summary['newborn_domain'] else 'low'
                    },
                    {
                        'name': 'punycode',
                        'risk': None if summary['punycode'] else 'low'
                    },
                    {
                        'name': 'suspicious_cookie',
                        'risk': None if summary['suspicious_cookie'] else 'low'
                    }
                )

                # get the list of subdomains and related domains
                for tech in report['data']['technologies']:
                    technologies.add(
                        {
                            'name': tech['name'],
                            'version': tech['version'],
                            'date': ''
                        }
                    )


    elif type == 'ip':
        url = config['api']['criminal_ip']['url_ip'].format(bacon)
        results = url_opener('GET', url, '', '', '', 'json', 'Criminal IP API')[0]

        # if the API call returns data
        if results['status'] == 200:
            ip_data['vulnerabilities'].add(
                {
                    'name': 'is_mobile',
                    'risk': None if results['is_mobile'] else 'info'
                },
                {
                    'name': 'is_hosting',
                    'risk': None if results['is_hosting'] else 'info'
                },
                {
                    'name': 'is_proxy',
                    'risk': None if results['is_proxy'] else 'medium'
                },
                {
                    'name': 'is_cloud',
                    'risk': None if results['is_cloud'] else 'info'
                },
                {
                    'name': 'is_tor',
                    'risk': None if results['is_tor'] else 'medium'
                },
                {
                    'name': 'is_vpn',
                    'risk': None if results['is_vpn'] else 'medium'
                },
                {
                    'name': 'is_darkweb',
                    'risk': None if results['is_darkweb'] else 'medium'
                },
                {
                    'name': 'is_scanner',
                    'risk': None if results['is_scanner'] else 'medium'
                },
                {
                    'name': 'is_snort',
                    'risk': None if results['is_snort'] else 'info'
                }
            )

            # iterate over ports found by Criminal IP
            for port in range (0, results['port']['count'] - 1):
                ip_data['ports'].add(
                    {
                        'port': port['open_port_no'],
                        'status': port['port_status'],
                        'protocol': port['socket'],
                        'service': port['protocol'],
                    }
                )
    else:
        errors = [f'Criminal IP API does not support "{type}".']
        error_printer(True, errors)

    # return gathered data
    return [
        ip_data,
        subdomains,
        related_domains,
        technologies,
        vulnerabilities
    ]



#!/usr/bin/env python

from typing import Dict
from bs4 import BeautifulSoup
import re


def hashable_finder(soup: BeautifulSoup, directive: str) -> Dict:
    """
    This function generates hash (SHA-256, SHA-384, or SHA-512 -> Base64) that is used by
    CSP directives 'child-src', 'connect-src', 'default-src', 'font-src', 'frame-src', 'img-src',
    'manifest-src', 'media-src', 'object-src', 'prefetch-src', 'script-src', 'script-src-elem',
    'script-src-attr', 'style-src', 'style-src-elem', 'style-src-attr', 'worker-src'.

    The hashing mechanism works like below:
    echo -n <TERM> | openssl sha256 -binary | openssl base64

    for more information read more on below links:
    - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy/Sources
    - https://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy
    - https://w3c.github.io/webappsec-csp

    Args:
        soup (BeautifulSoup): the BeautifulSoup format of the HTML page
        directive (src): the name of the directive

    Returns:
        hashables (dict): a dictionary contains all unique contents that can be hashed
    """
    # a variable to store list of hashable elements, attributes, and events
    hashables = {
        'base-uri': set(),
        'child-src': set(),
        'connect-src': set(),
        'default-src': set(),
        'font-src': set(),
        'frame-src': set(),
        'img-src': set(),
        'manifest-src': set(),
        'media-src': set(),
        'object-src': set(),
        'prefetch-src': set(),
        'script-src': set(),
        'script-src-elem': set(),
        'script-src-attr': set(),
        'style-src': set(),
        'style-src-elem': set(),
        'style-src-attr': set(),
        'worker-src': set(),
    }
    # some variables to store styles and scripts in BS4 tag and text formats
    styles = []    # BS4 tag format
    styles_ = []   # text format
    scripts = []   # BS4 tag format
    scripts_ = []  # text format

    # list of JavaScript events that can be used in any HTML element
    # i.e., <button onclick='{JavaScript}'>
    javascript_events = [
        "onafterprint",
        "onafterscriptexecute",
        "onanimationcancel",
        "onanimationend",
        "onanimationiteration",
        "onanimationstart",
        "onauxclick",
        "onbeforecopy",
        "onbeforecut",
        "onbeforeprint",
        "onbeforescriptexecute",
        "onbeforeunload",
        "onbegin",
        "onblur",
        "onbounce",
        "oncanplay",
        "oncanplaythrough",
        "onchange",
        "onclick",
        "onclose",
        "oncontextmenu",
        "oncopy",
        "oncuechange",
        "oncut",
        "ondblclick",
        "ondrag",
        "ondragend",
        "ondragenter",
        "ondragleave",
        "ondragover",
        "ondragstart",
        "ondrop",
        "ondurationchange",
        "onend",
        "onended",
        "onerror",
        "onfocusin",
        "onfocusout",
        "onfullscreenchange",
        "onhashchange",
        "oninput",
        "oninvalid",
        "onkeydown",
        "onkeypress",
        "onkeyup",
        "onload",
        "onloadeddata",
        "onloadedmetadata",
        "onloadend",
        "onloadstart",
        "onmessage",
        "onmousedown",
        "onmouseenter",
        "onmouseleave",
        "onmousemove",
        "onmouseout",
        "onmouseover",
        "onmouseup",
        "onmousewheel",
        "onmozfullscreenchange",
        "onpagehide",
        "onpageshow",
        "onpaste",
        "onpause",
        "onplay",
        "onplaying",
        "onpointerdown",
        "onpointerenter",
        "onpointerleave",
        "onpointermove",
        "onpointerout",
        "onpointerover",
        "onpointerrawupdate",
        "onpointerup",
        "onpopstate",
        "onprogress",
        "onrepeat",
        "onreset",
        "onresize",
        "onscroll",
        "onsearch",
        "onseeked",
        "onseeking",
        "onselect",
        "onselectionchange",
        "onselectstart",
        "onshow",
        "onstart",
        "onsubmit",
        "ontoggle",
        "ontouchend",
        "ontouchmove",
        "ontouchstart",
        "ontransitioncancel",
        "ontransitionend",
        "ontransitionrun",
        "ontransitionstart",
        "onunhandledrejection",
        "onunload",
        "onvolumechange",
        "onwebkitanimationend",
        "onwebkitanimationiteration",
        "onwebkitanimationstart",
        "onwebkittransitionend",
        "onwheel",
    ]

    # get the content of attributes of <base href='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/base-uri
    if directive == 'base-uri':
        hashables[directive].update(set(el['href'] for el in soup.findAll('base', href=True, nonce=False)))

    # get data related to <ifram> and related JavaScript APIs
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/child-src
    if directive in ['default-src', 'child-src']:
        # get the content of attributes of <iframe src='{URL}'> elements
        hashables[directive].update(set(el['src'] for el in soup.findAll('iframe', src=True, nonce=False)))
        # get <script> element contents
        scripts = soup.findAll('script', src=False, nonce=False) if not scripts else []
        # convert <script> element contents to string
        scripts_ = ' '.join([str(script) for script in scripts]) if not scripts_ else []
        # look for links in selected JavaScript APIs and add them to the set
        # the 're.DOTALL' flag is used to make the . character match newlines as well
        regex = r'Worker\s*\(["|\'](.*?)["|\']\)'
        hashables[directive].update(set(re.findall(regex, scripts_, re.DOTALL | re.IGNORECASE)))

    # get the content of attributes of <a ping='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/connect-src
    if directive in ['default-src', 'connect-src']:
        hashables[directive].update(set(el['ping'] for el in soup.findAll('a', ping=True, nonce=False)))

    # get the content of attributes of <style>@font-face {src: url('{URL}'); }</script> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/font-src
    if directive in ['default-src', 'font-src']:
        # get <style> element contents
        styles = soup.findAll('style', nonce=False) if not styles else []
        # convert <style> element contents to string
        styles_ = ' '.join([str(style) for style in styles]) if not styles_ else []
        # look for 'url's in '@import' rule and add them to the set
        # the 're.DOTALL' flag is used to make the . character match newlines as well
        regex = r'@font-face\s*.*?src\s*:\s*url\s*\(["|\'](.*?)["|\']\)'
        hashables[directive].update(set(re.findall(regex, styles_, re.DOTALL | re.IGNORECASE)))

    # get the content of attributes of <form action='{JavaSctript}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/form-action
    if directive in ['default-src', 'form-action']:
        hashables[directive].update(set(el['action'] for el in soup.findAll('form', action=True, nonce=False)))

    # get the content of attributes of <iframe src='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/frame-src
    if directive in ['default-src', 'frame-src']:
        hashables[directive].update(set(el['src'] for el in soup.findAll('iframe', src=True, nonce=False)))

    # get the content of attributes of <img src='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/img-src
    if directive in ['default-src', 'img-src']:
        hashables[directive].update(set(el['src'] for el in soup.findAll('img', src=True, nonce=False)))

    # get the content of attributes of <link rel='manifest' href='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/manifest-src
    if directive in ['default-src', 'manifest-src']:
        hashables[directive].update(set(el['href'] for el in soup.findAll('link', href=True,
                                                                          rel='manifest', nonce=False)))

    # get the content of attributes of <audio src='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/media-src
    if directive in ['default-src', 'media-src']:
        hashables[directive].update(set(el['src'] for el in soup.findAll(['audio', 'video', 'track'],
                                                                         src=True, nonce=False)))

    # get the content of attributes of <embed src='{URL}'>, <object data='{URL}'>, or <applet archive='{URL}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/object-src
    if directive in ['default-src', 'media-src']:
        hashables[directive].update(set(el['src'] for el in soup.findAll('embed', src=True, nonce=False)))
        hashables[directive].update(set(el['data'] for el in soup.findAll('object', data=True, nonce=False)))
        hashables[directive].update(set(el['archive'] for el in soup.findAll('applet', archive=True, nonce=False)))

    # get the content of all <script>{JavaScript}<script> elements and events
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/script-src
    if directive in ['default-src', 'script-src']:
        # get <script> element contents and add them to the set
        scripts = soup.findAll('script', src=False, nonce=False) if not scripts else []
        hashables[directive].update(set(el.text for el in scripts))

    # iterate over JavaScript events
    # events like <button onclick="send_form();">Send</button>
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/script-src-attr
    if directive in ['default-src', 'script-src-attr']:
        # find any HTML element contains the event: hashable in soup.select(f'[{event}]')
        hashables[directive].update(set(el[event] for event in javascript_events
                                        for el in soup.select(f'[{event}]')))

    # get the content of attributes of <script src='{url}'> elements
    # read more: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy/script-src-elem
    if directive in ['default-src', 'script-src-elem']:
        hashables[directive].update(set(el['src'] for el in soup.findAll('script', src=True, nonce=False)))

    # add the content of all <style>{CSS}</style> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/style-src
    if directive in ['default-src', 'style-src']:
        # get <style> element contents
        styles = soup.findAll('style', nonce=False) if not styles else []
        hashables[directive].update(set(el.text for el in styles))

    # add the content of attributes of <p style='{CSS}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/style-src-attr
    if directive in ['default-src', 'style-src-attr']:
        hashables[directive].update(set(el['style'] for el in soup.select('[style]')))

    # add the content of attributes of <link rel='stylesheet' href='{url}'> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/style-src-elem
    if directive in ['default-src', 'style-src-elem']:
        # add the content of attributes of <link rel='stylesheet' href='{url}'> elements
        hashables[directive].update(set(el['href'] for el in soup.findAll('link', href=True,
                                                                          rel='stylesheet', nonce=False)))
        # get <style> element contents
        styles = soup.findAll('style', nonce=False) if not styles else []
        # convert <style> element contents to string
        styles_ = ' '.join([str(style) for style in styles]) if not styles_ else []
        # look for 'url's in '@import' rule and add them to the set
        # the 're.DOTALL' flag is used to make the . character match newlines as well
        regex = r'@import\s+url\s*\(["|\'](.*?)["|\']\)'
        hashables[directive].update(set(re.findall(regex, styles_, re.DOTALL | re.IGNORECASE)))

    # get the content of attributes of <script>Worker('{URL}')</script>, <script>SharedWorker('{URL}')</script>, and
    # <script>serviceWorker.register('{URL}')</script> elements
    # read more: http://udn.realityripple.com/docs/Web/HTTP/Headers/Content-Security-Policy/worker-src
    if directive in ['default-src', 'worker-src']:
        # get <script> element contents
        scripts = soup.findAll('script', src=False, nonce=False) if not scripts else []
        # convert <script> element contents to string
        scripts_ = ' '.join([str(script) for script in scripts]) if not scripts_ else []
        # look for links in selected JavaScript APIs and add them to the set
        # the 're.DOTALL' flag is used to make the . character match newlines as well
        regex = r'[Worker|SharedWorker|serviceWorker\.register]\s*\(["|\'](.*?)["|\']\)'
        hashables[directive].update(set(re.findall(regex, scripts_, re.DOTALL | re.IGNORECASE)))

    # return results
    return hashables



#!/usr/bin/env python

from datetime import datetime
from usp.tree import sitemap_tree_for_homepage

from config import config


# get the raw date and its format, then covert it to a date object and print it
# in defined format
def date_formatter(date, date_format):
    # if the input 'date' is a string date not a date object
    if date_format:
        return datetime.strptime(date, date_format).strftime(config['date_format'])
    # if the input 'date' is a date object
    else:
        return date.strftime(config['date_format'])


# a function to return the value of a JSON key if it exists
def json_value(json, key):
    value = ''
    if key in json:
        if json[key] == 'false':
            value = False
        elif json[key] == 'true':
            value = True
        elif json[key] in ('null', 'none', None):
            value = ''
        else:
            value = json[key]

    # return value
    return value


# get the json and check if the key exist
# it is a recursive function that can check any values given in the list of
# keys in keys
def json_key_checker(json, keys):
    if keys[0] in json:
        if len(json) > 1 and keys[1] != '' and keys[1] in json[keys[0]]:
            keys.pop(0)
            json_key_checker(json[keys[0]], keys)
        else:
            return json[keys[0]]
    else:
        return ''


# a sub-function to parse the result of the regex find_all function
def re_position(term, pos):
    return term[pos] if (len(term) > pos) else ''


# print on the terminal (STDOUT) and the file
# all_prints = ''
# def printer(term):
#     # load the global variable
#     global all_prints

#     # print the term on the terminal (STDOUT)
#     print(term)

#     # strip styles from the term to be able to save it into a file
#     term = term.replace('\x1B[0m', '').replace('\x1B[30m', '').replace('\x1B[31m', '')
#     term = term.replace('\x1B[32m', '').replace('\x1B[33m', '').replace('\x1B[34m', '')
#     term = term.replace('\x1B[35m', '').replace('\x1B[36m', '').replace('\x1B[37m', '')
#     term = term.replace('\x1B[38m', '').replace('\x1B[39m', '').replace('\x1B[40m', '')
#     term = term.replace('\x1B[41m', '').replace('\x1B[42m', '').replace('\x1B[43m', '')
#     term = term.replace('\x1B[44m', '').replace('\x1B[45m', '').replace('\x1B[46m', '')
#     term = term.replace('\x1B[47m', '').replace('\x1B[48m', '').replace('\x1B[49m', '')
#     term = term.replace('\x1B[50m', '')

#     # save the term in a global variable to save in a file later
#     all_prints += term + '\n'

# sanitize the domain name and strip schema from it
# def domain_sanitizer(domain):
#     try:
#         # use urlparse() to split the URL into its component parts
#         parsed_url = urlparse(domain)

#         # check if netloc is empty (i.e., domain does not contain schema)
#         if not parsed_url.netloc:
#             domain = 'http://' + domain
#             parsed_url = urlparse(domain)

#         # form the components of the URL
#         url_components = [
#             parsed_url.netloc.lower(),
#             parsed_url.path,
#             parsed_url.params,
#             parsed_url.query,
#             parsed_url.fragment
#         ]

#     except Exception as error:
#         # handle any exceptions that may be raised
#         texts = [
#             f'Error in parsing the domain "{domain}".',
#             error
#         ]
#         print_error('exception', texts)
#         url_components = ['', '', '', '', '']

#     # return sanitized domain
#     return url_components


# get the details of the exception
# def exception_report():
#     # get error details
#     exception_type, exception_value, exception_trace = sys.exc_info()
#     # get the file location
#     file_name = os.path.split(exception_trace.tb_frame.f_code.co_filename)[1]
#     # get the line number
#     line_no = exception_trace.tb_lineno
#     # get the trace back to the source
#     exception_trace = ''.join(traceback.format_tb(exception_trace))
#     # return results
#     return [
#         str(exception_type),
#         str(exception_value),
#         str(exception_trace),
#         str(line_no),
#         str(file_name)
#     ]


# print the errors related to the exceptions, based on verbosity
# flag  = "exception" or True to print based on verbosity, and
#         False to print nothing as the error
# texts = ["text for verbosity 1", ..., "text for verbosity 5"]
# def print_error(flag, texts):
#     verbosity_1 = '      │        ├──🛑 '
#     verbosity_2 = '      │        │     '
#     verbosity_3 = '      │        │     '
#     verbosity_4 = '      │        │     '
#     verbosity_5 = '      │        │     '

#     # if flag is True/exception and verbosity is not 0 (or False)
#     if flag and config['verbosity']:
#         # print error messages for verbosity 1 (-v)
#         if config['verbosity'] >= 1 and len(texts) >= 1 and texts[0]:
#             printer(verbosity_1 + Fore.RED + texts[0] + Style.RESET_ALL)
#         # print error messages for verbosity 2 (-vv)
#         if config['verbosity'] >= 2 and len(texts) >= 2 and texts[1]:
#             printer(verbosity_2 + Fore.RED + texts[1] + Style.RESET_ALL)
#         # print error messages for verbosity 3 (-vvv)
#         if config['verbosity'] >= 3 and len(texts) >= 3 and texts[2]:
#             printer(verbosity_3 + Fore.RED + texts[2] + Style.RESET_ALL)
#         # print error messages for verbosity 4 (-vvvv)
#         if config['verbosity'] >= 4 and len(texts) >= 4 and texts[3]:
#             printer(verbosity_4 + Fore.RED + texts[3] + Style.RESET_ALL)
#         # print error messages for verbosity 5 (-vvvvv)
#         if config['verbosity'] >= 5:
#             if len(texts) >= 5 and texts[4]:
#                 printer(verbosity_5 + Fore.RED + texts[4] + Style.RESET_ALL)
#             # in case, it is an exception error message
#             if flag == 'exception':
#                 ex = exception_report()
#                 printer(
#                     verbosity_5 + 'Exception Message: ' + Fore.MAGENTA +
#                     ex[0] + ' ➜ ' + ex[4] + ':' + ex[3] + Fore.RED + ' == ' +
#                     ex[2] + ' ⚊ ' + ex[1] + Style.RESET_ALL
#                 )


# get the internal links based on the sitemap
# import logging
# def load_sitemap(domain):
    # a variable to store internal pages
    links = list()

    # disable logging of USP
    logging.getLogger('usp').disabled = True
    logging.getLogger('usp.helpers').disabled = True
    logging.getLogger('usp.fetch_parse').disabled = True
    logging.getLogger('usp.tree').disabled = True

    # get the sitemap
    sitemap = sitemap_tree_for_homepage(f'http://{domain}')

    # generate the internal page links
    for page in sitemap.all_pages():
        links.append(page.url)

    # return the results
    return links


# send the HTTP request based on the method and get the results
# in text or JSON format
# def run_requests(method, url, cookies, data, headers, type, name):
#     # some variables to store results
#     results = list()
#     history = list()
#     http_headers = list()
#     version = ''
#     status_code = 0

#     # print the subtitle based on the the variable "name"
#     if config['verbosity'] >= 2 and name:
#         printer(
#             '      │        ├□ ' + Fore.GREEN +
#             f'{name} is calling' + Style.RESET_ALL
#         )

#     # form the header and add a random User-Agent
#     if not headers:
#         headers = {'User-Agent': random.choice(config['user_agents'])}
#     else:
#         headers['User-Agent'] = random.choice(config['user_agents'])

#     # send the HTTP request based on the given method
#     try:
#         # run the GET request
#         if method == 'GET':
#             request = requests.get(
#                 url,
#                 headers=headers,
#                 timeout=config['delay']['requests_timeout']
#             )
#         # run the POST request
#         elif method == 'POST':
#             request = requests.post(
#                 url,
#                 cookies=cookies,
#                 data=json.dumps(data),
#                 headers=headers,
#                 timeout=config['delay']['requests_timeout']
#             )

#         # get the status code
#         status_code = int(request.status_code)

#         # get the history
#         history = request.history

#         # get the headers
#         http_headers = request.headers

#         # get the HTTP version
#         version = request.raw.version

#         # get the results if there is the type is either 'text', 'xml', or 'json'
#         if type == 'json':
#             results = json.loads(request.text)
#         elif type == 'text' or type == 'xml':
#             results = request.text
#         else:
#             results = request

#     # exceptions
#     except requests.exceptions.SSLError:
#         texts = [
#             f'The URL {url} cannot be opened due to an SSL error.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except requests.exceptions.TooManyRedirects:
#         texts = [
#             f'The provided {name} URL does not seem correct.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except requests.exceptions.HTTPError:
#         texts = [
#             f'Error in loading the {name} URL page.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except requests.exceptions.ConnectionError:
#         texts = [
#             f'Error in establishing the connection to the {name} URL.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except requests.exceptions.Timeout:
#         texts = [
#             f'Timeout error while opening {name} URL.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except requests.exceptions.RequestException:
#         texts = [
#             f'Error in reading the data from the {name} URL.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except json.decoder.JSONDecodeError:
#         texts = [
#             f'Error in reading the JSON data retrieved from the {name} call.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except ValueError:
#         texts = [
#             f'No Result is found or there was an error for {name}.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)
#     except Exception:
#         texts = [
#             f'An unknown error is ocurred while running {name}.',
#             '',
#             f'Status Code: {status_code}  -  API call URL: {url}'
#         ]
#         print_error('exception', texts)

#     # return results
#     finally:
#         return [
#             results,
#             status_code,
#             history,
#             http_headers,
#             version
#         ]



#!/usr/bin/env python

from time import sleep
from colorama import Fore, Style

from config import config
from modules.utils import json_value
from modules.utilities.url_opener import url_opener
from modules.utilities.printer import printer


def ssl_labs(domain):
    """
    ### SSL Certificates: SSL Labs

    This function uses the SSL Labs API to get the certificate's details.
    It, first, checks if there is any existing details. If there is any, loads
    it, if not, schedules the analysis. It also gets vulnerability assessments
    done by SSL Labs.
    # Read more: https://github.com/ssllabs/ssllabs-scan/blob/master/ssllabs-api-docs-v3.md

    # Input:  - a single domain name
    # Output: - a dictionary contains SSL certificate details
    """
    # a variable to store cert info
    cert_info = dict()
    # set the print arguments for the function "except_error_print"
    print_args = [True, '      │        ├──■ ', '      │        ├──■■ ']

    url = config['api']['ssl_labs']['url_status'].format(domain)

    # get the certificate from SSL Labs
    printer('      │        ├□ ' + Fore.GREEN + f'SSL Labs API is calling ({url})' + Style.RESET_ALL)

    # check if the certificate is in SSL Labs database
    # if not, run the scan and wait
    while True:
        # get the status
        json_data = url_opener('GET', url, '', '', '', 'json', 'Initial SSL Labs API')[0]
        status = json_data['status']

        # check the status
        if status in ('DNS', 'IN_PROGRESS'):
            printer('      │        ├──■ ' + Fore.YELLOW + f'Parsing SSL Certificate is not ready ({status})' + Style.RESET_ALL)
            printer('      │        └──■■ ' + Fore.YELLOW + json_data['statusMessage'] + Style.RESET_ALL)
            # wait for a certain time
            sleep(config['delay']['ssl_labs'])
            continue
        elif status == 'READY':
            break
        elif status == 'ERROR':
            printer('      │        ├──■ ' + Fore.RED + f'Error in parsing SSL Certificate ({status})' + Style.RESET_ALL)
            printer('      │        └──■■ Error: ' + Fore.RED + json_data['statusMessage'] + Style.RESET_ALL)
            return cert_info

    url = config['api']['ssl_labs']['url_detail'].format(domain, status['endpoints'][0]['ipAddress'])
    results = url_opener('GET', url, '', '', '', 'json', 'SSL Labs API')[0]['details']

    # continue only if there is any result
    if results:
        #initiate the result
        cert_info['control'] = dict()
        cert_info['vulnerability'] = dict()
        cert_info['control']['protocol'] = list()
        cert_info['control']['suite'] = list()

        cert_info['server_signature'] = json_value(results, 'serverSignature')

        cert_info['control']['grade'] = status['grade']

        # iterate over the suites and store them in a list
        for suites in results['suites']:
            cert_info['control']['protocol'].append(suites['protocol'])
            cert_info['control']['suite'][suites['protocol']] = list()
            for suite in suites['list']:
                # get the suite security flag
                flag = config['api']['ssl_labs']['suite_flag'][suite['q']] if suite['q'] else ''
                cert_info['control']['suite'][results['protocol']].append(flag + suite['name'])

        # get security controls out of certificate
        cert_info['control']['www_reachable'] = json_value(results, 'prefixDelegation')
        cert_info['control']['non_www_reachable'] = json_value(results, 'nonPrefixDelegation')
        cert_info['control']['compression'] = json_value(results, 'compressionMethods')
        cert_info['control']['npn'] = json_value(results, 'npnProtocols') if json_value(results, 'supportsNpn') else False
        cert_info['control']['alpn'] = json_value(results, 'alpnProtocols') if json_value(results, 'supportsAlpn') else False
        cert_info['control']['session_ticket'] = config['api']['ssl_labs']['session_resumption'][json_value(results, 'sessionTickets')]
        cert_info['control']['ocsp_stapling'] = json_value(results, 'ocspStapling')
        cert_info['control']['sni_required'] = json_value(results, 'sniRequired')
        cert_info['control']['rc4'] = json_value(results, 'supportsRc4')
        cert_info['control']['logjam'] = json_value(results, 'logjam')
        cert_info['control']['dhYsReuse'] = json_value(results, 'dhYsReuse')

        # get any vulnerabilities out of certificate
        cert_info['vulnerability']['beast'] = json_value(results, 'vulnBeast')
        cert_info['vulnerability']['heart_bleed'] = json_value(results, 'heartbleed')
        cert_info['vulnerability']['heartbeat'] = json_value(results, 'heartbeat')
        cert_info['vulnerability']['poodle'] = json_value(results, 'poodle')
        cert_info['vulnerability']['freak'] = json_value(results, 'freak')
        cert_info['vulnerability']['drown'] = json_value(results, 'drownVulnerable')
        cert_info['vulnerability']['drown_host'] = json_value(results, 'drownHosts')
        cert_info['vulnerability']['ecdh_parameter_reuse'] = json_value(results, 'ecdhParameterReuse')
        cert_info['vulnerability']['renegotiation'] = config['api']['ssl_labs']['renegotiation'][json_value(results, 'renegSupport')]
        cert_info['vulnerability']['session_resumption'] = config['api']['ssl_labs']['session_resumption'][json_value(results, 'sessionResumption')]
        cert_info['vulnerability']['openssl_ccs'] = config['api']['ssl_labs']['openssl_ccs'][json_value(results, 'openSslCcs')]
        cert_info['vulnerability']['openssl_lucky_m20'] = config['api']['ssl_labs']['openssl_lucky_m20'][json_value(results, 'openSSLLuckyMinus20')]
        cert_info['vulnerability']['ticket_bleed'] = config['api']['ssl_labs']['ticket_bleed'][json_value(results, 'ticketbleed')]
        cert_info['vulnerability']['bleichenbacher'] = config['api']['ssl_labs']['bleichenbacher'][json_value(results, 'bleichenbacher')]
        cert_info['vulnerability']['zombie_poodle'] = config['api']['ssl_labs']['zombie_poodle'][json_value(results, 'zombiePoodle')]
        cert_info['vulnerability']['golden_doodle'] = config['api']['ssl_labs']['golden_doodle'][json_value(results, 'goldenDoodle')]
        cert_info['vulnerability']['zero_length_padding_oracle'] = config['api']['ssl_labs']['zero_length_padding_oracle'][json_value(results, 'zeroLengthPaddingOracle')]
        cert_info['vulnerability']['sleeping_poodle'] = config['api']['ssl_labs']['sleeping_poodle'][json_value(results, 'sleepingPoodle')]
        cert_info['vulnerability']['poodle_tls'] = config['api']['ssl_labs']['poodle_tls'][json_value(results, 'poodleTls')]

    return cert_info



#!/usr/bin/env python

from typing import List, Dict
from bs4 import BeautifulSoup


def meta_parser(soup: BeautifulSoup) -> List[Dict[str, str]]:
    """
    Extracts meta data tags from the given BeautifulSoup object and returns a
    list of dictionaries containing the name and value attributes.

    Args:
        soup (BeautifulSoup): A BeautifulSoup object representing an HTML page.

    Returns:
        List[Dict[str, str]]: A list of dictionaries containing the extracted meta data.
    """
    # A list to store the extracted meta data
    html_meta_tags = []

    # Find "meta" tags within the HTML content
    # Ignore the ones has 'http-equiv' instead of 'name'
    meta_tags = soup.findAll('meta', {'http-equiv': None})

    # Iterate over the found metadata and extract the name and value attributes
    for tag in meta_tags:
        # Set the attribute type
        if tag.get('name'):
            attribute_type = 'name'
        elif tag.get('http-equiv'):
            attribute_type = 'http-equiv'
        elif tag.get('charset'):
            attribute_type = 'charset'
        elif tag.get('property'):
            attribute_type = 'property'
        elif tag.get('itemprop'):
            attribute_type = 'itemprop'
        else:
            print(f'Cannot parse the HTML Meta tag: {tag}')
            continue

        name = tag.get(attribute_type).lower()
        value = tag.get('content', '')

        html_meta_tags.append({
            'attribute': attribute_type,
            'name': name if attribute_type != 'charset' else 'charset',
            'value': value if attribute_type != 'charset' else name,
        })

    # Return results
    return html_meta_tags



#!/usr/bin/env python

from colorama import Fore, Back, Style

from config import config
from modules.utilities.printer import printer


# get the wohis of the domain form different sources
def whois_lookup(domain, api):
    # some variables
    i = 0

    # print findings
    printer('      │\n      ├───' + Fore.BLACK + Back.WHITE + ' Whois ' + Style.RESET_ALL)

    # call the asked whois API
    if api == 'whoisxml':
        whois, flag = whoisxml(whois, domain)
    elif api == 'whoxy':
        whois, flag = whoxy(whois, domain)
    else:
        printer(Fore.RED + '      │      ■ ' + Fore.RED + 'Whois API is not defined; read the help.' + Style.RESET_ALL)
        return whois

    # print the result on STDOUT
    if flag:
        printer('      │      ■ Registration Date:      ' + Fore.YELLOW + whois['create_date'] + Style.RESET_ALL)
        printer('      │      ■ Last Modification Date: ' + Fore.YELLOW + whois['update_date'] + Style.RESET_ALL)
        printer('      │      ■ Expiration Date:        ' + Fore.YELLOW + whois['expiration_date'] + Style.RESET_ALL)
        printer('      │      ■ Domain Age:             ' + Fore.YELLOW + str(whois['domain_age_days']) + ' Days' + Style.RESET_ALL)
        if config['verbosity'] > 2:
            printer('      │      ' + Fore.CYAN + '■ Registrar Details' + Style.RESET_ALL)
            printer('      └┐      ■■  Name:                ' + Fore.YELLOW + whois['registrar']['name'] + Style.RESET_ALL)
            printer('       │      ■■  IANA ID:             ' + Fore.YELLOW + str(whois['registrar']['iana_id']) + Style.RESET_ALL)
            printer('       │      ■■  Website:             ' + Fore.YELLOW + whois['registrar']['website'] + Style.RESET_ALL)
            printer('       │      ■■  Wohis Server:        ' + Fore.YELLOW + whois['registrar']['whois_server'] + Style.RESET_ALL)
            printer('       │      ■■  Email:               ' + Fore.YELLOW + whois['registrar']['email'] + Style.RESET_ALL)
            printer('      ┌┘      ■■  Phone:               ' + Fore.YELLOW + whois['registrar']['phone'] + Style.RESET_ALL)
            printer('      │      ' + Fore.CYAN + '■ Registrant Details' + Style.RESET_ALL)
            printer('      └┐      ■■  Name:                ' + Fore.YELLOW + whois['registrant']['name'] + Style.RESET_ALL)
            printer('       │      ■■  Country:             ' + Fore.YELLOW + whois['registrant']['country'] + Style.RESET_ALL)
            printer('       │      ■■  Email:               ' + Fore.YELLOW + whois['registrant']['email'] + Style.RESET_ALL)
            printer('      ┌┘      ■■  Phone:               ' + Fore.YELLOW + whois['registrant']['phone'] + Style.RESET_ALL)
            printer('      │      ' + Fore.CYAN + '■ Administrative Details' + Style.RESET_ALL)
            printer('      └┐      ■■  Name:                ' + Fore.YELLOW + whois['administrative']['name'] + Style.RESET_ALL)
            printer('       │      ■■  Country:             ' + Fore.YELLOW + whois['administrative']['country'] + Style.RESET_ALL)
            printer('       │      ■■  Email:               ' + Fore.YELLOW + whois['administrative']['email'] + Style.RESET_ALL)
            printer('      ┌┘      ■■  Phone:               ' + Fore.YELLOW + whois['administrative']['phone'] + Style.RESET_ALL)
            printer('      │      ' + Fore.CYAN + '■ Technical Details' + Style.RESET_ALL)
            printer('      └┐      ■■  Name:                ' + Fore.YELLOW + whois['technical']['name'] + Style.RESET_ALL)
            printer('       │      ■■  Country:             ' + Fore.YELLOW + whois['technical']['country'] + Style.RESET_ALL)
            printer('       │      ■■  Email:               ' + Fore.YELLOW + whois['technical']['email'] + Style.RESET_ALL)
            printer('      ┌┘      ■■  Phone:               ' + Fore.YELLOW + whois['technical']['phone'] + Style.RESET_ALL)
        if config['verbosity'] >= 2:
            printer('      │      ' + Fore.CYAN + '■ Name Servers' + Style.RESET_ALL)
            for ns in whois['name_servers']:
                i += 1
                if i == 1:
                    printer('      └┐      ■■  Nameserver {0}:         '.format(i) + Fore.YELLOW + ns + Style.RESET_ALL)
                elif i == len(whois['name_servers']):
                    printer('      ┌┘      ■■  Nameserver {0}:         '.format(i) + Fore.YELLOW + ns + Style.RESET_ALL)
                else:
                    printer('       │      ■■  Nameserver {0}:         '.format(i) + Fore.YELLOW + ns + Style.RESET_ALL)

    else:
        if config['verbosity'] >= 4:
            printer('      │      ■■ Error: ' + Fore.RED + 'No valid Whois API key for "{0}" is set.'.format(api) + Style.RESET_ALL)
        return whois

    # return the result in the format of list instead of set
    return whois



#!/usr/bin/env python

from typing import List, Dict
from colorama import Fore, Style
from bs4 import BeautifulSoup

from modules.utilities.printer import printer
from modules.http_header.nonce_finder import nonce_finder
from modules.http_header.nonce_validator import nonce_validator
from modules.http_header.hashable_finder import hashable_finder
from modules.http_header.hash_validator import hash_validator


def csp_parser(csp: str, soup: BeautifulSoup) -> List[Dict[str, List[str]]]:
    """
    Parse the directives and sources of the HTTP Header "content-security-policy".
    The CSP is in this format: "directive1 source11 source12; directive2 source21 source22 source23;"

    Args:
        csp (str): The value of the CSP header.
        soup (BeautifulSoup): the BeautifulSoup format of the HTML page

    Returns:
        A list of dictionaries containing the directive and sources.

    Example:
        >>> csp_parser("default-src 'self'; img-src https://*")
        [{'directive': 'default-src', 'sources': ["'self'"]},
         {'directive': 'img-src', 'sources': ['https://*']}]
    """
    # a variable to store list of hashable elements, attributes, and events
    hashables = {
        'base-uri': set(),
        'child-src': set(),
        'connect-src': set(),
        'default-src': set(),
        'font-src': set(),
        'frame-src': set(),
        'img-src': set(),
        'manifest-src': set(),
        'media-src': set(),
        'object-src': set(),
        'prefetch-src': set(),
        'script-src': set(),
        'script-src-elem': set(),
        'script-src-attr': set(),
        'style-src': set(),
        'style-src-elem': set(),
        'style-src-attr': set(),
        'worker-src': set(),
    }
    # a dictionary to store data
    CSPs = []
    # two variables to store nonces and hashes
    nonces = []
    hashes = []
    # for each of the directives contains '-src' that are absent, the user agent looks for the 'default-src'
    # this variable assures that finding hashes for 'default-src' doesn't stop looking for other directives
    default_src_directive = False

    # get directives after sanitizing the CSP header
    csp = ' '.join(csp.split())
    directives = csp.strip().replace('; ', ';').replace(' ;', ';').split(';')

    # if 'default-src' is one of the directives bring it to the beginning of the the list
    # with this, we assure that it will be parsed at the first and would not mess with other directives
    if 'default_src' in directives:
        # flag it
        default_src_directive = True
        # bring 'default_src' to the position 0 in the list; so we'll have ['default_src', ...]
        directives = directives.insert(0, directives.pop(directives.index('default_src')))

    # iterate over the CSP directives
    for directive_source in directives:
        # sanitize each CSP value
        directive_source = directive_source.strip().replace('"', '').replace('\'', '')
        filter(None, directive_source)

        # get the directive and sources from the value
        directive = directive_source.split(' ')[0].strip()
        sources = directive_source.split(' ')[1:]

        # sanitize the list of sources
        filter(None, sources)

        # remove the nonce value
        # nonce is a random value that would be generated per each page load
        # so, it is not unique. read more: https://web.dev/strict-csp/
        for i, source in enumerate(sources):
            # check for nonces
            if source.lower().startswith('nonce-'):
                # find nonces in in-line JavaScript and CSS Style codes
                if not nonces:
                    nonces = nonce_finder(soup)

                # validate nonces
                sources[i] = nonce_validator(source, nonces)

            # check for hashes
            elif source.lower().startswith('sha'):
                # if the directive is 'default_src', get all data
                if directive == 'default_src':
                    hashables = hashable_finder(soup, directive)
                # if the directive is not parsed previously and if it is not 'default-src'
                elif not hashables[directive] or default_src_directive:
                    # get the hashable elements, events, and attributes
                    hashables[directive].update(hashable_finder(soup, directive)[directive])

                # validate hashes
                sources[i], hashable, hash = hash_validator(source, hashables[directive])

                # remove the hashed element from the list to prevent hashing elements multiple times
                if hashable in hashables[directive]:
                    hashables[directive].remove(hashable)

                # add the hash value to a list
                if hash:
                    hashes.append(hash)

        # continue only if there is a directive
        if directive:
            # write the results in the output
            CSPs.append(
                {
                    'directive': directive,
                    'sources': sources
                }
            )

            # print the results
            printer(f'      │          ∘ {directive + ":":20}{Fore.YELLOW}{sources}{Style.RESET_ALL}')

    # return results
    return [
        CSPs,
        nonces,
        hashes
    ]



#!/usr/bin/env python

from datetime import datetime
import OpenSSL
import ssl

from config import config
from modules.utils import date_formatter
from modules.subdomain.utils import sub_related_domains
from modules.utilities.error_printer import error_printer


def openssl(domain):
    """
    This function get the certificate details from Python OpenSSL module (pyOpenSSL).
    It first gets the list of all certificates, sort them by ID, and load the
    latest certificate (based on ID). Afterwards, parse the page using RE.

    # Input:  - a single domain name
    # Output: - a dictionary contains SSL certificate details
    """

    # A variable to store cert info
    cert_info = dict()
    # Set the print arguments for the function "except_error_print"
    print_args = [True, '      │        ├──■ ', '      │        ├──■■ ']

    try:
        # Get the certificate
        cert = ssl.get_server_certificate((domain, 443))
        # Parse the certificate
        x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, cert)

        # Iterate over the certificate extensions to get alternative names
        for i in range(0, x509.get_extension_count()):
            extension = x509.get_extension(i)
            if 'subjectAltName' in str(extension.get_short_name()):
                alt_names = str(extension).split(',')
                sub, alt = sub_related_domains(alt_names, domain)
                cert_info['subdomains'].update(sub)
                cert_info['related_domains'].update(alt)

        # Parse cert info and write into a dictionary
        cert_info['expired'] = x509.has_expired()
        cert_info['serial_number'] = str(x509.get_serial_number())
        cert_info['version'] = x509.get_version()
        cert_info['signature'] = x509.get_signature_algorithm()
        cert_info['validity']['issue_date'] = date_formatter(x509.get_notBefore().decode('ascii'),
                                                             '%Y%m%d%H%M%SZ')
        cert_info['validity']['expiration_date'] = date_formatter(x509.get_notAfter().decode('ascii'),
                                                                  '%Y%m%d%H%M%SZ')
        cert_info['validity']['days'] = (
            cert_info['validity']['expiration_date'] - cert_info['validity']['issue_date']).days
        cert_info['validity']['past_days'] = (datetime.now() - datetime.strptime(cert_info['validity']['issue_date'],
                                                                                 config['date_format'])).days
        cert_info['validity']['left_days'] = (datetime.strptime(cert_info['validity']['expiration_date'],
                                                                config['date_format']) - datetime.now()).days
        cert_info['fingerprint']['md5'] = x509.digest('md5').decode()
        cert_info['fingerprint']['sha1'] = x509.digest('sha1').decode()
        cert_info['fingerprint']['sha256'] = x509.digest('sha256').decode()
        cert_info['fingerprint']['sha512'] = x509.digest('sha512').decode()
        cert_info['issuer']['common_name'] = x509.get_issuer().CN
        cert_info['issuer']['organization_name'] = x509.get_issuer().O
        cert_info['issuer']['organization_unit_name'] = x509.get_issuer().OU
        cert_info['issuer']['country'] = x509.get_issuer().C
        cert_info['issuer']['state'] = x509.get_issuer().ST
        cert_info['issuer']['city'] = x509.get_issuer().L
        cert_info['issuer']['email_address'] = x509.get_issuer().emailAddress
        cert_info['subject']['common_name'] = x509.get_subject().CN
        cert_info['subject']['organization_name'] = x509.get_subject().O
        cert_info['subject']['organization_unit_name'] = x509.get_subject().OU
        cert_info['subject']['country'] = x509.get_subject().C
        cert_info['subject']['state'] = x509.get_subject().ST
        cert_info['subject']['city'] = x509.get_subject().L
        cert_info['subject']['email_address'] = x509.get_subject().emailAddress

    except ssl.SSLError as err:
        args = [print_args, f'There is an error in certificate of the host "{domain}" is failed.', err, '']
        error_printer(True, args)
    except Exception:
        args = [print_args, f'Error in getting or parsing certificate for the host "{domain}".', '', '']
        error_printer(True, args)

    # Return the result
    return cert_info



#!/usr/bin/env python


"""
    ### Vulnerability Scanner: Larger.io API

    This function gets the list of the technologies (services) used in the given
    domain name.

    Read more: - https://www.larger.io/user/api

    # Input:  - a single domain name or IP address
              - the scan type which would be deep or quick
    # Output: - a set contains emails related to the given bacon
              - a set of dictionaries contains technologies used in the bacon
"""


from time import sleep

from config import config
from modules.utils import url_opener, error_printer, json_key_checker, printer


def larger_io(bacon, scan_type):
    # variables to store results
    emails = set()
    technologies = set()

    # get the type of the scan; quick or deep
    # if it is defined by STDIN, the setting from the config file will be ignored
    if scan_type == '':
        scan_type = config['scan_type']['technology']

    # get the delay defined in the config file
    delay = config['delay']['large_io']

    # sleep for a certain time as the Large.io API has limits
    sleep(delay)

    # get the results in JSON from Larger.io based on the scan type
    api_key = config['api']['larger_io']['api_key']
    if scan_type == 'deep':
        url = config['api']['larger_io']['url_slow'].format(api_key, bacon)
    else:
        url = config['api']['larger_io']['url_fast'].format(api_key, bacon)
    results = url_opener('GET', url, '', '', '', 'json', 'Larger.io API')[0]

    # if the API call returns data
    if 'status' in results and results['status'] != 'false':
        # get the list of technologies
        if 'app' in results:
            for tech in results['apps']:
                technologies.add(
                    {
                        'name': json_key_checker(tech, ['name']),
                        'version': json_key_checker(tech, ['name']),
                    }
                )

        # get the list of emails
        if 'emails' in results:
            emails = json_key_checker(tech, ['emails'])

    else:
        errors = [
            f'There is an error in getting the data from Larger.io',
            '',
            f'{results["message"]}',
            '',
            ''
        ]
        error_printer(True, errors)

    # return gathered data
    return [
        technologies,
        emails
    ]



#!/usr/bin/env python

from bs4 import BeautifulSoup

from config import config
from modules.dns.subdomain_takeover import subdomain_takeover
from modules.utilities.url_opener import url_opener


def dns_spy(domain):
    """
        ### DNS Records: DNS Spy API

        This function opens the DNS Spy page and parse the DNS records gathered by
        DNS-Spy page.

        # Input:  - a single domain name
        # Output: - a dictionary contains DNS records details
    """
    # initial variables
    dns_records = {}
    subdomains = list()

    # form the URL and the header
    url = config['api']['dns_spy']['url'].format(domain)

    # open the URL
    results = url_opener('GET', url, '', '', '', 'text', 'DNS Spy API')
    text_data = results[0]
    status_code = results[1]

    # if it returns Error 404 which means there is no data for the domain
    if str(status_code) == '404':
        return [
            dns_records,
            set(subdomains)
        ]

    # parse the page
    rows = BeautifulSoup(text_data, "html.parser").find('table', {'id': 'domain-table'}).find_all('tr')
    for row in rows:
        # temp variable
        temp = list()

        if 'Record' in str(row):
            continue
        row = row.find_all("td")
        # sanitize the data
        row[0] = row[0].get_text().replace(' ', '').replace('\n', '')
        row[1] = row[1].get_text().replace(' ', '').replace('\n', '')
        row[2] = row[2].get_text().replace(' ', '').replace('\n', '')
        row[3] = row[3].get_text().replace('"', '').replace('\\', '')
        row[3] = row[3].replace('                                                  ', '')
        row[3] = row[3].replace('                                               ', '')
        row[3] = row[3].replace('                                              ', '')

        if row[1] == domain and row[0] != 'TXT':
            value = row[3].split('\n')
            # add values to the results and remove duplicates
            temp += list(set(value))
        # find subdomains A\AAAA records
        elif row[0] in ('A', 'AAAA'):
            subdomains.append(row[1].split(domain)[0])
        else:
            sdt_answer = ''
            # ignore the followings
            if row[0] in 'TXT' and row[1].startswith('_dmarc'):
                continue
            # find subdomains based on CNAME
            if row[0] in 'CNAME':
                subdomains.append(row[1].split('.' + domain)[0])
                # if the record is CNAME, check for the subdomain takeover
                sdt = subdomain_takeover(row[3])
                # if vulnerable to subdomain takeover
                if sdt == 1:
                    sdt_answer = '(Vuln_Subdomain_Takeover) '
            row[3] = row[3].replace('\n', ' ')
            value = sdt_answer + row[1] + ' ' + row[2] + ' ' + row[3]
            # add values to the results
            temp.append(value)

        # add the data to results and remove duplicates
        if row[0] in dns_records:
            dns_records[row[0]] += list(set(temp))
        else:
            dns_records[row[0]] = list(set(temp))

        # remove empty items
        if '' in dns_records[row[0]]:
            dns_records[row[0]].remove('')

    # return results
    return [
        dns_records,
        set(subdomains)
    ]



#!/usr/bin/env python

from modules.dns.dns_resolver import dns_resolver
from modules.dns.utils import dns_record_type, rr_record_finder


def dns_dig(domain, key, record):
    """
    ### DNS Checker: DNS Dig

    This function gets the value of all DNS records.

    # Input:  - a single domain name
              - the key, subdomain, or selector.subdomain
              - a single DNS record either in an RR type (int) or case-insensitive string; e.g. 1 or A or a
    # Output: - a set of dictionaries contains a single DNS record details
    """
    # A variable to store results
    results = set()

    # Get the RR type of the DNS record
    rr_record = rr_record_finder(record)

    # Resolve the DNS record
    answers = dns_resolver(domain, record)

    # Iterate over answers
    for answer in answers:
        results.add({
            'record': record,
            'type': dns_record_type(rr_record, key, answer),
            'key': key,
            'value': answer.to_text()
        })

    # Return the list of the DNS records
    return results



#!/usr/bin/env python

import re
from colorama import Fore, Style

from config import config
from modules.utils import re_position
from modules.utils import date_formatter
from modules.utilities.printer import printer
from modules.utilities.url_opener import url_opener


def crt_sh(domain):
    """
    ### SSL Certificates: crt.sh API

    This function get the latest certificate details from crt.sh.
    It first gets the list of all certificates, sort them by ID, and load the
    latest certificate (based on ID). Afterwards, parse the page using RE.

    # Input:  - a single domain name
    # Output: - a dictionary contains SSL certificate details
    """
    # a variable to store cert info
    cert_info = dict()
    # get the date format of the crt.sh from the config file
    date_format = config['api']['crt_sh']['date_format']

    # download the certificate page on CRT
    url = config['api']['crt_sh']['url_all'].format(domain)
    printer('      │        ├□ ' + Fore.GREEN + f'CRT.sh API is calling ({url})' + Style.RESET_ALL)
    results = url_opener('GET', url, '', '', '', 'json', 'Crt.sh API')[0]

    # continue only if there is any result
    if results:
        # sort certificates by ID to get the details of the latest one
        results = sorted(results, key=lambda k: k['id'], reverse=True)

        url = config['api']['crt_sh']['url_single'].format(results[0]['id'])
        cert = url_opener('GET', url, '', '', '', 'text', 'Crt.sh API')[0]

        # fix the HTML format of the space
        cert = cert.replace('&nbsp;', ' ')

        # get the info of the latest certificate
        cert_info['signature'] = re_position(re.findall(r"Signature Algorithm[ :]*(.*?)<BR>", cert), 0)
        cert_info['serial_number'] = re_position(re.findall(r"Serial Number[ :]<\/A><BR>[ =]*(.*?)<BR>", cert), 0)
        cert_info['validity']['issue_date'] = date_formatter(re_position(re.findall(r"Not Before[ =:]*(.*?)<BR>", cert), 0), date_format)
        cert_info['validity']['expiration_date'] = date_formatter(re_position(re.findall(r"Not After[ =:]*(.*?)<BR>", cert), 0), date_format)
        cert_info['fingerprint']['sha256'] = re_position(re.findall(r"search.censys.io/certificates/(.*?)\">", cert), 0)
        cert_info['issuer']['common_name'] = re_position(re.findall(r"commonName[ =]*(.*?)<BR>", cert), 0)
        cert_info['issuer']['organization_name'] = re_position(re.findall(r"organizationName[ =]*(.*?)<BR>", cert), 0)
        cert_info['issuer']['organization_unit_name'] = re_position(re.findall(r"organizationalUnitName[ =]*(.*?)<BR>", cert), 0)
        cert_info['issuer']['country'] = re_position(re.findall(r"countryName[ =]*(.*?)<BR>", cert), 0)
        cert_info['issuer']['state'] = re_position(re.findall(r"stateOrProvinceName[ =]*(.*?)<BR>", cert), 0)
        cert_info['issuer']['city'] = re_position(re.findall(r"localityName[ =]*(.*?)<BR>", cert), 0)
        cert_info['subject']['common_name'] = re_position(re.findall(r"commonName[ =]*(.*?)<BR>", cert), 1)
        cert_info['subject']['organization_name'] = re_position(re.findall(r"organizationName[ =]*(.*?)<BR>", cert), 1)
        cert_info['subject']['organization_unit_name'] = re_position(re.findall(r"organizationalUnitName[ =]*(.*?)<BR>", cert), 1)
        cert_info['subject']['country'] = re_position(re.findall(r"countryName[ =]*(.*?)<BR>", cert), 1)
        cert_info['subject']['state'] = re_position(re.findall(r"stateOrProvinceName[ =]*(.*?)<BR>", cert), 1)
        cert_info['subject']['city'] = re_position(re.findall(r"localityName[ =]*(.*?)<BR>", cert), 1)
        cert_info['control']['lint'] = re.findall(r"<SPAN[\sclass=\"notice|warning|fatal|error\"]*>\s*(.*?)\s*</SPAN>", cert)

    return cert_info



#!/usr/bin/env python

from typing import List
import logging
from colorama import Fore, Style
from usp.tree import sitemap_tree_for_homepage
import usp.exceptions

from config import config
from modules.utilities.printer import printer
from modules.utilities.exception_handling import exception_details


def load_sitemap(hostname: str) -> List[str]:
    """
    Fetches the sitemap for the given hostname and returns a list of internal links.

    Args:
        hostname (str): A string representing the hostname.

    Returns:
        A list of internal page links.
    """
    # A variable to store retrieved pages from the sitemap
    pages = []

    # Disable logging of USP
    logging.getLogger('usp').disabled = True
    logging.getLogger('usp.helpers').disabled = True
    logging.getLogger('usp.fetch_parse').disabled = True
    logging.getLogger('usp.tree').disabled = True

    # Get the sitemap
    try:
        sitemap_tree = sitemap_tree_for_homepage(f'http://{hostname}')

        # Generate the internal page links
        pages = [page.url for page in sitemap_tree.all_pages()]

    except usp.exceptions.GunzipException:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in in I/O (input/output).{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
                    f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')

    except usp.exceptions.SitemapException:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Can\'t run further, e.g. wrong input parameters.{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
                    f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')

    except usp.exceptions.SitemapXMLParsingException:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}XML parsing exception to be handled gracefully.{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
                    f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')

    except usp.exceptions.StripURLToHomepageException:
        ex = exception_details()
        printer(f'      │ ■ {Fore.RED}Error in in I/O (input/output).{Style.RESET_ALL}')
        if config['verbosity'] >= 4:
            printer(f'      │ ■■ ERROR: {Fore.MAGENTA}{ex[0]} → {ex[4]}:{ex[3]} ({ex[3].errno}, {ex[3].strerror}) '
                    f'{Fore.RED}  {ex[1]}{Style.RESET_ALL}')

    finally:
        pages += hostname

    # Return the results
    return pages
